# Performance Measurement
This document describes how POD can be measured in terms of load capability.

# Prerequisites
Make sure you do testing among implementations on the same DB file! Querying bigger DB might be significantly slower than small one.

## Sqlite cannot open database error
This error was visible while running `locust` with 1000 concurrent "users", task that reaches "search" endpoint
for 10 different pod users.

Reason for that is POD process under heavy load exceed max FD default capacity that is on linux set to 1024 FD's.

One pod user is 8 pool connections (number of cores) * 3 database FDs (sqlite db, wal file, shm file) + at least one TCP connection from client(s)

`locust` itself reserves around 1k FD for TCP connections leaving very few room for FDs mentioned above, resulting in SQLITE_CANTOPEN (14) error.
Solution is to extend number of FDs, probably in docker container used in production env

## Increase limit of opened FDs
Check the limit:
```
ulimit -Sn
ulimit -Hn
```
Change the limit:
```
vim /etc/security/limits.conf

username    soft nofile     10000
```

Log out, log in


# Commands:
## Apache Benchmark
Apache Benchmark can be used to see throughput on one endpoint for one POD user. Useful, as a first step in performance analysis.

Example search:
```
ab -kc 1000 -n 10000  -t 'application/json' -p ~/memri/curl_helpers/search_sample.json http://127.0.0.1:3030/v4/$OWNER_KEY/search
```
Request:
```
{
    "auth": {
      "type": "ClientAuth",
      "databaseKey": "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb"
    },
    "payload": {
      "dateServerModified>=": 0,
      "[[edges]]": {},
      "_limit": 100,
      "_sortOrder": "Desc"
    }
}
```

Example write:
```
ab -kc 1000 -n 10000 -t 'application/json' -p /home/szym/memri/curl_helpers/create_fake_message.json  http://localhost:3030/v4/$OWNER_KEY/create_item

```

Request:
```
{
  "auth": {
      "type": "ClientAuth",
      "databaseKey": "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb"
  },
  "payload": {
    "type": "Message",
    "content": "this is a fake content"
  }
}
```

#  Locust
[Locust](https://docs.locust.io/en/stable/index.html) allow us to simulate consumer load on the POD.
Under `pod/perf_testing` directory there is a `locustfile.py` that Contains 2 clients, one simulating Frontend operation, second simulating Plugin behavior. Each simulated POD owner has those 2 clients (can be think of as one POD owner has one tab opened in the browser + one running plugin). In order to run it:
* [Optional] Set the `POD_DB_DIR` env variable to point to the directory where databases reside.
* Set number of locust `--users` to 2 * POD owners you want. Locust naming here is really unfortunate and it's easy to mess up locust users with POD users.

So following example will run 40 POD users (hence 40 * 2 = 80 locust `users`). Their number will rise gradually by 5 (2*5 = 10).
```
POD_DB_DIR=~/memri/pod/data/db locust --users 80 --spawn-rate 10 -H http://localhost:3030
```

In the Web UI you will see POD owners called `BEEFXXXX` and clients doing things on their behalf.

Locust allows to tweak number of tasks per client, number of clients and their probability to be spawned, request periodicity, or even simulate [non-linear loads](https://docs.locust.io/en/stable/writing-a-locustfile.html#wait-time)

# Current state
POD uses SQLite database with WAL mode enabled. That give us non blocking readers, and one writer. Concurrent writers has to be serialized, and it's done by starting transaction with ```BEGIN EXCLUSIVE``` statement.

There is a connection pool per each POD owner containing number connections equal to count of logical CPU cores. This is also equal to number of async threads that are provided by tokio runtime.

Connections from the pool are re-used across requests. There is always exactly one connection in use by exactly one thread.

All SQLite operations are done by `rusqlite` library that is not async. It turned out that operations even though are blocking in nature, they are so fast that moving them to separate thread by using ```spawn_blocking``` makes no sense. The context switching takes longer than DB execution itself.

This however might be a matter of change, see [future steps](#future-steps) section.

## SQLX
There was and attempt to use ```sqlx``` library to make code asynchronous. It was abandoned for reasons:
1. `sqlx` performance [is a joke](https://github.com/diesel-rs/metrics). The best performance shows the library we currently use - `rusqlite`
2. It requires to rewrite almost everything in the POD since the TX object type is changing from rusqlite to sqlx. POD does not have interface abstraction between logic and database driver. Rusqlite is tightly coupled in the codebase.

# Future steps
Possible improvement ideas when tackling this problem once again:

* Instead of blocking ```pool.get()```, create ```thread_local``` hash map with connections - one connection per POD owner on each async thread?
* Current implementation suffers by ```N+1``` query problem a lot.
* Looks like the current implementation of dynamic schema in the DB is very inefficient and that should be a main topic to focus on in regards of further perf improvements.
* SQLite itself might be tweaked using opening flags, or pragmas, there is very cool article that shows how to select best parameters for given system
https://ericdraken.com/sqlite-performance-testing/

  Current implementation is very close to conclusions made in there
* Add heuristic: if request makes significant changes and that will cost in time more than acceptable on async thread, offload it to separate thread -> ```spawn_blocking().await```. How to check if request should be offloaded?
  * Check body request size (would work for write transactions), if above some threshold -> offload.
  * In write transactions: set sql_busy timeout to small value (or find a way to do try_write), if tx fails, offload to the thread with longer timeout - not sure if good idea, will spam the system with threads, but will unblock async workers
  * What about readers? Check the query complexity just before passing it to the DB, and decide on that?
* Find use for `PRAGMA thread`??
* Add periodic tasks that use pragmas:
  * `vaccum` will shorten DB in size if it's possible,
  * `optimize` improve sql query execution,
  * `wal_checkpoint = RESTART` That will force the `-wal` file be committed to the database, this one is expensive, requires exclusive lock on the whole database.
  * Fiddle with `journal_size`, now it's set to 10MB, value taken from thin air. Maybe there are better choices?

