use crate::error::{ErrorContext, Result};
use crate::plugin_auth_crypto::DatabaseKey;
use crate::{constants, database_migrate_refinery};
use rusqlite::{Connection, OpenFlags};
use std::collections::HashMap;
use std::path::{Path, PathBuf};
use std::sync::Arc;
use tokio::sync::RwLock;

pub type DBConnection = rusqlite::Connection;

/// Keeps track on database initialization for POD owners, and creates connections
#[derive(Debug, Clone)]
pub struct DBConnectionManager {
    init_db: Arc<RwLock<HashMap<String, Arc<RwLock<PathBuf>>>>>,
}

impl DBConnectionManager {
    pub fn new() -> Self {
        Self {
            init_db: Arc::new(RwLock::new(HashMap::new())),
        }
    }

    pub async fn initialize_db(
        &self,
        owner: &str,
        database_key: &DatabaseKey,
    ) -> Result<Connection> {
        // Double locking pattern, all that juggling between locks is done
        // for one purpose, block the POD as less as possible.

        // This is the happy path, all expect first call goes here
        {
            let db = self.init_db.read().await;
            if let Some(database_path) = db.get(owner) {
                let database_path = database_path.read().await;
                return self.make_connection(database_key, &*database_path);
            }
        };

        let conn = {
            // There is no database yet.
            let mut db = self.init_db.write().await;

            // There might be situation someone else put it moment ago, so check for existence again
            if let Some(database_path) = db.get(owner) {
                let database_path = database_path.read().await;

                self.make_connection(database_key, &*database_path)?
            } else {
                // Under exclusive write lock there is nothing in the hash map, create the path then
                let database_path = format!("{}{}", &owner, constants::DATABASE_SUFFIX);
                let database_path = PathBuf::from(constants::DATABASE_DIR).join(database_path);

                let owner_entry = Arc::new(RwLock::new(database_path.clone()));
                db.insert(owner.to_string(), owner_entry.clone());

                // Gather exclusive lock on owner's entry
                let _owner_write_guard = owner_entry.write().await;

                // Drop the init_db write lock, so other requests
                // from other owners can access it and move forward
                drop(db);

                // Do migration under the exclusive owner-lock, so it's guaranteed all connections will see a database
                // with migrations in place.
                let mut conn = self.make_connection(database_key, &database_path)?;

                // Do migration, that might be time consuming, delegate on separate task
                tokio::task::spawn_blocking(move || database_migrate_refinery::migrate(&mut conn))
                    .await??;

                // Finally return a connection
                self.make_connection(database_key, &database_path)?
            }
        };

        Ok(conn)
    }

    fn make_connection<P>(&self, database_key: &DatabaseKey, database_path: P) -> Result<Connection>
    where
        P: AsRef<Path>,
    {
        let conn = rusqlite::Connection::open_with_flags(
            database_path,
            // Create the file if not exist
            OpenFlags::SQLITE_OPEN_CREATE
            // Open in "multi thread" mode, sqlite disables synchronization in connection.
            // That is correct as long as one connection is used by exactly one thread at a time,
            // which is the case for the POD
            | OpenFlags::SQLITE_OPEN_NO_MUTEX
            // Open for reading and writing
            | OpenFlags::SQLITE_OPEN_READ_WRITE
            // Use private cache per connection, shared cache introduces a lot of synchronization
            // between them impacting performance seriously.
            | OpenFlags::SQLITE_OPEN_PRIVATE_CACHE,
        )?;

        DatabaseKey::execute_sqlite_pragma(database_key, &conn).context_str(
            "Failed to open the database file (did databaseKey change between runs?)",
        )?;

        conn.execute_batch(
    "
            PRAGMA foreign_keys = ON;

            -- Strategy for syncing data to the storage, NORMAL is recommended for WAL
            PRAGMA synchronous = NORMAL;

            -- Temporal storage set to MEMORY, don't use FILE setting, since SQLCipher does not do encryption on them
            PRAGMA temp_store = 2;

            -- Milliseconds that connection will wait to be unblocked with its operations, after that period SQL_BUSY is returned
            PRAGMA busy_timeout = 3000;

            -- Set the maximum number of pages in the database file.
            -- Currently 1GB given a page size of 4096 (default >= 3.12.0). 1024 * 1024 * 1024 / 4096 = 262144
            PRAGMA max_page_count = 262144;

            -- Use Write Ahead Logging journal that allows multiple readers and at most one writer to work seamlessly
            PRAGMA journal_mode = WAL;

            -- Do auto-checkpoints if journal file exceeds this number of pages
            -- The checkpoint is PASSIVE, meaning will succeed iff there is nothing preventing from doing it
            PRAGMA wal_autocheckpoint = 1000;

            -- The DB will truncate the WAL journal to that size after doing a checkpoint.
            -- This is a trade off made here, keep the journal on the storage for better performance
            -- (do not re-create it each time is needed), but keep it reasonably small so -wal files
            -- will not clutter our cloud space too much.
            -- This setting is a matter of change after releasing POD to the wild and check how it behaves
            -- on users accounts.
            -- Set to 10MB
            -- Note under heavy write load -wal may exceed this limit extensively, it will be shortened
            -- after the write 'wave' goes away.
            PRAGMA journal_size_limit = 10485760;

            -- Number of bytes of the database file that will be accessed using memory-mapped I/O.
            -- mmaping reduces amount of read/write syscalls.
            -- Write happens to -wal file, so there is no that much impact while writing
            -- 30 MB
            PRAGMA mmap_size = 31457280;

            --- Neg value to set Kibibytes (positive for pages)
            PRAGMA cache_size = -1280000;
        ",
        )?;

        Ok(conn)
    }
}

impl Default for DBConnectionManager {
    fn default() -> Self {
        Self::new()
    }
}

#[cfg(test)]
mod tests {
    use rusqlite::{
        ffi::{Error, ErrorCode},
        Error::SqliteFailure,
    };

    use super::*;

    struct TempDb {
        pub database_path: PathBuf,
    }

    impl TempDb {
        fn new(owner: &str) -> Self {
            let database_path = format!("{}{}", owner, constants::DATABASE_SUFFIX);
            let database_path = PathBuf::from(constants::DATABASE_DIR).join(database_path);

            if !Path::new(constants::DATABASE_DIR).is_dir() {
                std::fs::create_dir_all(constants::DATABASE_DIR).unwrap();
            }

            Self { database_path }
        }

        fn rm_files(&self) {
            let _ = std::fs::remove_file(&self.database_path);
            let _ = std::fs::remove_file(&format!("{}-shm", self.database_path.display()));
            let _ = std::fs::remove_file(&format!("{}-wal", self.database_path.display()));
        }
    }

    impl Drop for TempDb {
        fn drop(&mut self) {
            self.rm_files();
        }
    }

    const NORMAL: u32 = 1;
    const ON: u32 = 1;
    const MEMORY: u32 = 2;
    const BUSY_TIMEOUT: u32 = 3000;
    const WAL: &str = "wal";
    const MAX_PAGE_COUNT: u32 = 262144;

    #[tokio::test]
    async fn pragmas_are_set() {
        let owner = "testing_pragmas_are_set";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let db_manager = DBConnectionManager::new();

        let conn = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        assert_eq!(
            ON,
            conn.query_row("pragma foreign_keys;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );

        assert_eq!(
            NORMAL,
            conn.query_row("pragma synchronous;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );

        assert_eq!(
            MEMORY,
            conn.query_row("pragma temp_store;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );
        assert_eq!(
            BUSY_TIMEOUT,
            conn.query_row("pragma busy_timeout;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );
        assert_eq!(
            WAL,
            conn.query_row("pragma journal_mode;", [], |row| row
                .get::<usize, String>(0))
                .unwrap()
        );
        assert_eq!(
            MAX_PAGE_COUNT,
            conn.query_row("pragma max_page_count;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );
    }

    #[tokio::test]
    async fn fails_on_rollback_ok_on_wal() {
        let owner = "testing_fails_on_rollback_ok_on_wal";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let db_manager = DBConnectionManager::new();

        let connection_a = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        let connection_b = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        // Journal: this one gets called
        // connection_a.busy_handler(Some(|a| {
        //     println!("busy handler for connection_a called!");
        //     true
        // }))
        // .unwrap();

        connection_a
            .execute_batch(
                "
            create table if not exists t(x integer);
            insert into t values(1);
            insert into t values(2);
            ",
            )
            .unwrap();

        connection_a.execute_batch("begin transaction;").unwrap();

        connection_b.execute_batch("begin transaction;").unwrap();

        // In rollback: connection_b gets SHARED lock
        connection_b.execute_batch("select * from t;").unwrap();

        // In rollback: connection_a gets RESERVED (intent to write lock)
        connection_a
            .execute_batch("insert into t values(1);")
            .unwrap();

        let journal_mode = connection_a
            .query_row("pragma journal_mode;", [], |row| {
                row.get::<usize, String>(0)
            })
            .unwrap();

        if journal_mode == WAL {
            // In WAL: it's designed in such a way readers does not interfere with writers,
            // so this passes
            assert!(matches!(connection_a.execute_batch("commit;"), Ok(())));
        } else {
            // In rollback: tries to go to EXCLUSIVE, can get it, iff there are no other locks
            // but there is one SHARED held by connection_b, connection_a waits BUSY_TIMEOUT and returns with error
            // If there is busy_handler set, then sqlite will call it as long as it returns true,
            // or is able to proceed.
            assert!(matches!(
                connection_a.execute_batch("commit;"),
                Err(SqliteFailure(
                    Error {
                        code: ErrorCode::DatabaseBusy,
                        extended_code: 5
                    },
                    Some(_)
                ))
            ))
        }

        connection_b.execute_batch("commit;").unwrap();
    }

    #[test_log::test(tokio::test)]
    async fn deadlock_detection() {
        let owner = "testing_deadlock_detection";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let db_manager = DBConnectionManager::new();

        let connection_a = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        connection_a
            .busy_handler(Some(|_a| {
                println!("busy handler for connection_a called!");
                true
            }))
            .unwrap();

        let connection_b = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        connection_a
            .execute_batch(
                "
            create table if not exists t(x integer);
            insert into t values(1);
            insert into t values(2);
            ",
            )
            .unwrap();

        connection_a.execute_batch("begin transaction;").unwrap();

        connection_b.execute_batch("begin transaction;").unwrap();

        // connection_a gets SHARED lock
        connection_a.execute_batch("select * from t;").unwrap();

        // connection_b gets SHARED lock
        connection_b.execute_batch("select * from t;").unwrap();

        // connection_b gets RESERVED, there can be only 1 at a time of such kind
        connection_b
            .execute_batch("insert into t values(1);")
            .unwrap();

        // connection_a tries to get RESERVED. It fails IMMEDIATELY, because in order to jump to
        // EXCLUSIVE later, all locks has to be cleared out, but that will never happen,
        // because connection_b also holds RESERVED and also will wait to all locks to drop.
        // If both connection_a and connection_b would gather RESERVED, then:
        // connection_a would wait for connection_b to drop, connection_b would wait for connection_a
        // to drop - making a deadlock.
        // Such situation is detected by SQLITE, and aborts this call immediately ->
        // allowing connection connection_b to gather exclusive lock later.
        //
        // Note, no busy_handlers are called
        assert!(matches!(
            connection_a.execute_batch("insert into t values(1);"),
            Err(SqliteFailure(
                Error {
                    code: ErrorCode::DatabaseBusy,
                    extended_code: 5
                },
                // Result message says "database locked" that suggest SQLITE_LOCKED
                // error, which is completely different thing
                // and makes a lot of confusion.
                Some(_msg)
            ))
        ));

        connection_a.execute_batch("commit;").unwrap();

        connection_b.execute_batch("commit;").unwrap();
    }

    #[tokio::test]
    async fn reader_and_2_writers() {
        let owner = "testing_reader_and_2_writers";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let db_manager = DBConnectionManager::new();

        let reader_a = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        let writer_b = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();
        let writer_c = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        writer_b
            .execute_batch(
                "
            create table if not exists t(x integer);
            insert into t values(1);
            insert into t values(2);
            ",
            )
            .unwrap();

        reader_a.execute_batch("begin transaction;").unwrap();
        reader_a.execute_batch("select * from t;").unwrap();

        writer_b.execute_batch("begin transaction;").unwrap();
        writer_b.execute_batch("insert into t values(1);").unwrap();

        writer_c.execute_batch("begin transaction;").unwrap();

        // In journal and WAL
        // waits busy_timeout milliseconds for writer_A to finish
        assert!(matches!(
            writer_c.execute_batch("insert into t values(1);"),
            Err(SqliteFailure(
                Error {
                    code: ErrorCode::DatabaseBusy,
                    extended_code: 5
                },
                // Result message says "database locked" that suggest SQLITE_LOCKED
                // error, which is completely different thing
                // and makes a lot of confusion.
                Some(_msg)
            ))
        ));
    }

    #[tokio::test]
    async fn only_valid_key_is_honored() {
        let owner = "only_valid_key_is_honored";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let database_key_invalid = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbCD".to_string(),
        )
        .unwrap();

        let db_manager = DBConnectionManager::new();

        let conn = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();

        const ENTRY: i32 = 123454321;
        // Insert record to a table
        conn.execute_batch(&format!(
            "
            create table if not exists t(x integer);
            insert into t values({ENTRY});
            ",
        ))
        .unwrap();

        // Invalid key should return an error
        assert!(db_manager
            .initialize_db(owner, &database_key_invalid)
            .await
            .is_err());

        // Connect again with valid credentials, should be possible to read from db
        let conn = db_manager
            .initialize_db(owner, &database_key)
            .await
            .unwrap();
        let mut q = conn.prepare("select * from t").unwrap();
        let mut rows = q.query([]).unwrap();

        while let Ok(Some(row)) = rows.next() {
            assert_eq!(row.get::<_, i32>(0).unwrap(), ENTRY);
        }
    }
}
