use crate::api_model::SortOrder;
use graphql_parser::query::{
    parse_query, Definition, Field, OperationDefinition, Selection, SelectionSet, Text, Value,
};
use std::convert::TryFrom;
use warp::http::status::StatusCode;

use crate::error::{Error, Result};
use std::collections::HashMap;

#[derive(Debug)]
pub struct QueryASTNode {
    pub item_type: Option<String>,
    pub arguments: Arguments,
    pub selectors: Vec<QueryASTNode>,
    pub properties: Vec<String>,
    pub edges: HashMap<String, QueryASTNode>,
}

impl QueryASTNode {
    fn new() -> Self {
        QueryASTNode {
            item_type: None,
            arguments: Arguments::new(),
            selectors: Vec::new(),
            properties: Vec::new(),
            edges: HashMap::new(),
        }
    }
}

fn parse_query_recursive(selection_set: SelectionSet<String>) -> Result<QueryASTNode> {
    let mut result = QueryASTNode::new();
    for selection in selection_set.items {
        if let Selection::Field(field) = selection.clone() {
            if field_is_property(&field) {
                result.properties.push(field.name);
            } else {
                let mut edge_ast_node = parse_query_recursive(field.selection_set.clone())?;
                for (arg_name, arg_value) in field.arguments {
                    edge_ast_node.arguments.add_arg(arg_name, arg_value)?;
                }
                for selection in field.selection_set.items {
                    if let Selection::InlineFragment(fragment) = selection {
                        let mut fragment_node = parse_query_recursive(fragment.selection_set)?;
                        let cond =
                            format!("{}", fragment.type_condition.unwrap()).replace("on ", "");
                        fragment_node.item_type = Some(cond.clone());
                        edge_ast_node
                            .properties
                            .extend(fragment_node.properties.clone());
                        edge_ast_node.selectors.push(fragment_node);
                    }
                }
                result.edges.insert(field.name, edge_ast_node);
            }
        }
    }
    Ok(result)
}

pub fn parse_graphql_query(query_string: &str) -> Result<QueryASTNode> {
    // TODO Only returns first Query
    // TODO error handling
    let parsed = match parse_query::<String>(query_string) {
        Ok(res) => res,
        Err(err) => {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: format!("Invalid GraphQL query {}", err),
            });
        }
    };

    if parsed.definitions.len() > 1 {
        panic!("graphQL only supports queries");
    }

    let def = parsed.definitions.first().ok_or_else(|| Error {
        code: StatusCode::BAD_REQUEST,
        msg: "Attempted to parse empty query".to_owned(),
    })?;

    if let Definition::Operation(OperationDefinition::Query(top_level_query)) = def.to_owned() {
        if top_level_query.selection_set.items.len() > 1 {
            panic!("graphQL supports a single query");
        }
        // println!("{:}", top_level_query);
        let query = top_level_query
            .selection_set
            .items
            .first()
            .unwrap()
            .to_owned();
        if let Selection::Field(field) = query {
            let mut result = parse_query_recursive(field.selection_set)?;
            result.item_type = Some(field.name);
            for (arg_name, arg_value) in field.arguments {
                result.arguments.add_arg(arg_name, arg_value)?;
            }
            Ok(result)
        } else {
            panic!("Fragments are not supported")
        }
    } else {
        panic!("Selectionsets/Mutations/Subscriptions are not supported");
    }
}

pub fn field_is_property(field: &Field<String>) -> bool {
    field.selection_set.items.is_empty()
}

#[derive(Debug, Clone)]
#[allow(clippy::upper_case_acronyms)]
pub enum Filter {
    EQ {
        property: String,
        value: String,
    },
    NE {
        property: String,
        value: String,
    },
    LT {
        property: String,
        value: String,
    },
    GT {
        property: String,
        value: String,
    },
    GTE {
        property: String,
        value: String,
    },
    #[allow(clippy::upper_case_acronyms)]
    LTE {
        property: String,
        value: String,
    },
    // Box needed, cant do recursive memory allocation
    #[allow(clippy::upper_case_acronyms)]
    AND {
        left: Box<Filter>,
        right: Box<Filter>,
    },
    #[allow(clippy::upper_case_acronyms)]
    OR {
        left: Box<Filter>,
        right: Box<Filter>,
    },
}

impl Filter {
    pub fn to_sql_string(&self) -> String {
        match self {
            Filter::AND { left, right } => format!(
                "({:} AND {:})",
                &left.to_sql_string(),
                &right.to_sql_string()
            ),
            Filter::OR { left, right } => format!(
                "({:} OR {:})",
                &left.to_sql_string(),
                &right.to_sql_string()
            ),
            Filter::EQ { property, value } => Filter::format_eq(property, value),
            Filter::NE { property, value } => Filter::format_ne(property, value),
            Filter::LT { property, value } => format!("({:} < {:})", property, value),
            Filter::GT { property, value } => format!("({:} > {:})", property, value),
            Filter::LTE { property, value } => format!("({:} <= {:})", property, value),
            Filter::GTE { property, value } => format!("({:} >= {:})", property, value),
        }
    }

    fn format_eq(property: &String, value: &String) -> String {
        match value.to_lowercase().as_str() {
            "null" => format!("({:} is {:})", property, value),
            _ => format!("({:} = {:})", property, value),
        }
    }

    fn format_ne(property: &String, value: &String) -> String {
        match value.to_lowercase().as_str() {
            "null" => format!("({:} is not {:})", property, value),
            _ => format!("({:} != {:})", property, value),
        }
    }
}

#[derive(Debug)]
pub struct Arguments {
    pub limit: Option<u32>,
    pub offset: Option<u32>,
    pub sort_order: Option<SortOrder>,
    pub sort_property: Option<String>,
    pub filters: Option<Filter>,
}

impl Arguments {
    fn new() -> Arguments {
        Arguments {
            limit: None,
            filters: None,
            offset: None,
            sort_order: None,
            sort_property: None,
        }
    }

    fn add_arg<'a, T: Text<'a>>(
        &mut self,
        arg_name: String,
        arg_value: Value<'a, T>,
    ) -> Result<()> {
        match arg_name.as_str() {
            "limit" => self.set_limit(arg_value)?,
            "filter" => self.set_filter(arg_value)?,
            "offset" => self.set_offset(arg_value)?,
            "order_asc" => self.set_order(arg_value, SortOrder::Asc)?,
            "order_desc" => self.set_order(arg_value, SortOrder::Desc)?,
            _ => {
                return Err(Error {
                    code: StatusCode::BAD_REQUEST,
                    msg: format!("Unknown graphQL argument: {}", arg_name),
                })
            }
        }

        Ok(())
    }

    fn set_filter<'a, T: Text<'a>>(&mut self, arg_value: Value<'a, T>) -> Result<()> {
        if self.filters.is_some() {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "GraphQL error: multiple limits defined".to_owned(),
            });
        }
        pub fn parse_filter<'a, T: Text<'a>>(filter_ast: &Value<'a, T>) -> Result<Filter> {
            if let Value::Object(filter_tree) = filter_ast {
                if filter_tree.len() != 1 {
                    return Err(Error {
                        code: StatusCode::BAD_REQUEST,
                        msg: "filter ast can only contain one object on each level".to_owned(),
                    });
                }
                let (k, v) = filter_tree.iter().next().unwrap();
                let filter = match k.as_ref() {
                    "and" => {
                        if let Value::List(filters) = v {
                            if filters.len() != 2 {
                                return Err(Error {
                                    code: StatusCode::BAD_REQUEST,
                                    msg: "AND/OR filters should always have exactly 2 children"
                                        .to_owned(),
                                });
                            }
                            let left = Box::<Filter>::new(parse_filter(&filters[0])?);
                            // let left = Box::<Filter>::new(parse_filter(filters.as_ref()[0])?);
                            let right = Box::<Filter>::new(parse_filter(&filters[1])?);
                            Filter::AND { left, right }
                        } else {
                            return Err(Error {
                                code: StatusCode::BAD_REQUEST,
                                msg: "and filters requires a list of arguments".to_owned(),
                            });
                        }
                    }
                    "or" => {
                        if let Value::List(filters) = v {
                            if filters.len() != 2 {
                                return Err(Error {
                                    code: StatusCode::BAD_REQUEST,
                                    msg: "AND/OR filters should always have exactly 2 children"
                                        .to_owned(),
                                });
                            }
                            let left = Box::<Filter>::new(parse_filter(&filters[0])?);
                            let right = Box::<Filter>::new(parse_filter(&filters[1])?);
                            Filter::OR { left, right }
                        } else {
                            return Err(Error {
                                code: StatusCode::BAD_REQUEST,
                                msg: "\"or\" filters require a list of arguments".to_owned(),
                            });
                        }
                    }
                    // a nested filter
                    _ => {
                        let filter_property = k.as_ref();
                        if let Value::Object(filter_obj) = v {
                            if filter_obj.len() != 1 {
                                return Err(Error {
                                    code: StatusCode::BAD_REQUEST,
                                    msg: "normal (non AND/OR) filters can only contain one key value pair".to_owned(),
                                });
                            } else {
                                let (filter_name, filter_value) = filter_obj.iter().next().unwrap();
                                match filter_name.as_ref() {
                                    "eq" => Filter::EQ {
                                        property: filter_property.to_string(),
                                        value: filter_value.to_string(),
                                    },
                                    "ne" => Filter::NE {
                                        property: filter_property.to_string(),
                                        value: filter_value.to_string(),
                                    },
                                    "gt" => Filter::GT {
                                        property: filter_property.to_string(),
                                        value: filter_value.to_string(),
                                    },
                                    "lt" => Filter::LT {
                                        property: filter_property.to_string(),
                                        value: filter_value.to_string(),
                                    },
                                    "lte" => Filter::LTE {
                                        property: filter_property.to_string(),
                                        value: filter_value.to_string(),
                                    },
                                    "gte" => Filter::GTE {
                                        property: filter_property.to_string(),
                                        value: filter_value.to_string(),
                                    },
                                    _ => panic!(
                                        "could not parse filter type {:}",
                                        filter_name.as_ref()
                                    ),
                                }
                            }
                        } else {
                            return Err(Error {
                                code: StatusCode::BAD_REQUEST,
                                msg: format!("invalid graphQL filter {:}, json obj expected", v),
                            });
                        }
                    }
                };
                Ok(filter)
            } else {
                panic!("invalid graphQL filter");
            }
        }
        let filter = parse_filter(&arg_value)?;
        self.filters = Some(filter);
        Ok(())
    }

    fn set_limit<'a, T: Text<'a>>(&mut self, arg_value: Value<'a, T>) -> Result<()> {
        if self.limit.is_some() {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "GraphQL error: multiple limits defined".to_owned(),
            });
        }

        if let Value::Int(v) = arg_value {
            let v = u32::try_from(v.as_i64().unwrap())?;
            self.limit = Some(v);
            Ok(())
        } else {
            Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "Invalid GraphQL query limit".to_owned(),
            })
        }
    }

    fn set_offset<'a, T: Text<'a>>(&mut self, arg_value: Value<'a, T>) -> Result<()> {
        if self.offset.is_some() {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "GraphQL error: multiple offsets defined".to_owned(),
            });
        }

        if let Value::Int(v) = arg_value {
            let v = u32::try_from(v.as_i64().unwrap())?;
            self.offset = Some(v);
            Ok(())
        } else {
            Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "Invalid GraphQL query offset".to_owned(),
            })
        }
    }

    fn set_order<'a, T: Text<'a>>(
        &mut self,
        arg_value: Value<'a, T>,
        order: SortOrder,
    ) -> Result<()> {
        if self.sort_order.is_some() | self.sort_property.is_some() {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "GraphQL error: multiple orders defined".to_owned(),
            });
        }

        if let Value::Enum(o) = arg_value {
            self.sort_property = Some(o.as_ref().to_owned());
            self.sort_order = Some(order);
            Ok(())
        } else {
            Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: "Invalid GraphQL query order".to_owned(),
            })
        }
    }
}

#[cfg(test)]
pub mod tests {
    use super::*;
    use crate::error::Result;

    #[test]
    fn test_parse_query() -> Result<()> {
        let test_query = "
            query {
                Message (limit: 1000, offset: 200 order_asc: id) {
                    id
                }
            }
        ";

        let query_ast = parse_graphql_query(test_query)?;
        println!("{:?}", query_ast.arguments);

        Ok(())
    }

    #[test]
    fn test_filters() -> Result<()> {
        let test_query = "
            query {
                Message(filter: {title: {eq: \"GraphQL\"}}) {
                    id
                    content
                    author {
                        id
                    }
                }
            }
        ";

        let query_ast = parse_graphql_query(test_query)?;

        assert_eq!(query_ast.item_type, Some("Message".to_owned()));
        assert_eq!(query_ast.properties, ["id", "content"]);
        assert_eq!(query_ast.edges.len(), 1);

        let (edge_name, edge_ast) = query_ast.edges.iter().next().unwrap();
        assert_eq!(edge_name, "author");
        assert_eq!(edge_ast.properties, ["id"]);
        println!("{:?}", query_ast.arguments);

        Ok(())
    }

    #[test]
    fn test_filters2() -> Result<()> {
        let test_query = "
            query {
                Message(filter: {and: [{title: {eq: \"GraphQL\"}}, {name: {ne: \"notGraphQl\"}}]}) {
                    id
                    content
                    author {
                        id
                    }
                }
            }
        ";

        let query_ast = parse_graphql_query(test_query)?;

        assert_eq!(query_ast.item_type, Some("Message".to_owned()));
        assert_eq!(query_ast.properties, ["id", "content"]);
        assert_eq!(query_ast.edges.len(), 1);

        let (edge_name, edge_ast) = query_ast.edges.iter().next().unwrap();
        assert_eq!(edge_name, "author");
        assert_eq!(edge_ast.properties, ["id"]);
        println!("{:?}", query_ast.arguments);

        Ok(())
    }

    #[test]
    fn test_filters_invalid_queries() -> Result<()> {
        let test_query = "
            query {
                Message(filter: {and: [{title: {eq: \"GraphQL\"}}, {name: {ne: \"notGraphQl\"}}, {name: {ne: \"invalid\"}}]}) {
                    id
                    content
                    author {
                        id
                    }
                }
            }
        ";

        let query_ast = parse_graphql_query(test_query);
        assert!(query_ast.is_err());
        Ok(())
    }
}
