use std::{borrow::Cow, collections::HashMap, env};

use crate::error::{Error, Result};
use oauth::RequestBuilder;
pub use oauth_client as oauth;
use reqwest::{
    header::{HeaderName, HeaderValue},
    Client, Method, StatusCode,
};
use serde_json::json;
use std::convert::TryFrom;
use warp::http;

#[derive(Debug)]
pub struct AsyncRequestBuilder {
    inner: reqwest::RequestBuilder,
}

impl RequestBuilder for AsyncRequestBuilder {
    type ReturnValue = tokio::task::JoinHandle<reqwest::Result<String>>;
    type ClientBuilder = Client;

    fn new(method: Method, url: &'_ str, client: &Self::ClientBuilder) -> Self {
        Self {
            inner: client.clone().request(method, url),
        }
    }

    fn body(mut self, b: String) -> Self {
        self.inner = self.inner.body(b);

        self
    }

    fn header<K, V>(mut self, key: K, val: V) -> Self
    where
        HeaderName: TryFrom<K>,
        HeaderValue: TryFrom<V>,
        <HeaderName as TryFrom<K>>::Error: Into<http::Error>,
        <HeaderValue as TryFrom<V>>::Error: Into<http::Error>,
    {
        self.inner = self.inner.header(key, val);

        self
    }

    fn send(self) -> std::result::Result<Self::ReturnValue, oauth_client::Error> {
        Ok(tokio::spawn(async {
            self.inner.send().await?.error_for_status()?.text().await
        }))
    }
}

pub fn twitter_api_keys() -> (Result<String>, Result<String>) {
    let consumer_key = match env::var("TWITTER_APP_KEY") {
        Ok(val) => Ok(val),

        Err(_e) => Err(Error {
            code: StatusCode::BAD_REQUEST,
            msg: "TWITTER_APP_KEY env var is not set".to_string(),
        }),
    };
    let consumer_secret = match env::var("TWITTER_APP_SECRET") {
        Ok(val) => Ok(val),
        Err(_e) => Err(Error {
            code: StatusCode::BAD_REQUEST,
            msg: "TWITTER_APP_SECRET env var is not set".to_string(),
        }),
    };
    (consumer_key, consumer_secret)
}

pub async fn oauth1_access_token(
    oauth_token: String,
    oauth_verifier: String,
    oauth_token_secret: String,
) -> Result<serde_json::Value> {
    let (consumer_key, consumer_secret) = twitter_api_keys();

    let consumer = oauth_client::Token::new(consumer_key?, consumer_secret?);
    let token = oauth_client::Token::new(oauth_token, oauth_token_secret);

    let mut param = HashMap::new();
    param.insert(Cow::from("oauth_verifier"), Cow::from(oauth_verifier));

    let bytes = oauth::post::<AsyncRequestBuilder>(
        "https://api.twitter.com/oauth/access_token",
        &consumer,
        Some(&token),
        Some(&param),
        &reqwest::Client::new(),
    )?
    .await??;

    let resp = std::str::from_utf8(bytes.as_ref()).unwrap();

    let mut res_params: HashMap<String, String> = HashMap::new();
    for q in resp.split('&') {
        let (k, v) = q.split_once('=').unwrap();
        let _ = res_params.insert(k.into(), v.into());
    }
    let res = serde_json::to_value(&res_params).unwrap();
    Ok(res)
}

pub async fn oauth1_request_token(
    callback_url: String,
    service: String,
) -> Result<serde_json::Value> {
    println!("authenticating {}", service);
    if service == "twitter" {
        let (consumer_key, consumer_secret) = twitter_api_keys();

        let mut param = HashMap::new();
        param.insert(Cow::from("oauth_callback"), Cow::from(callback_url));

        let consumer = oauth_client::Token::new(consumer_key?, consumer_secret?);

        let bytes = oauth_client::post::<AsyncRequestBuilder>(
            "https://api.twitter.com/oauth/request_token",
            &consumer,
            None,
            Some(&param),
            &reqwest::Client::new(),
        )?
        .await??;

        let resp = std::str::from_utf8(bytes.as_ref()).unwrap();

        let mut res_params: HashMap<String, String> = HashMap::new();
        for q in resp.split('&') {
            let (k, v) = q.split_once('=').unwrap();
            let _ = res_params.insert(k.into(), v.into());
        }
        if !res_params.contains_key("oauth_token") {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: format!("response did not return oauth token: {}", resp),
            });
        }
        let res = json!({
            "oauth_token": res_params["oauth_token"],
            "oauth_token_secret": res_params["oauth_token_secret"],
            "oauth_callback_confirmed": res_params["oauth_callback_confirmed"]
        });
        Ok(res)
    } else {
        Err(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("unknown oauth service: {}", service),
        })
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[tokio::test]
    async fn test_oauth() {
        let runs_on_protected = env::var("CI_COMMIT_REF_PROTECTED");

        if matches!(runs_on_protected,  Ok(s) if s == "true") {
            let mut res = oauth1_request_token(
                "https://dev.app.memri.io/oauth".to_string(),
                "twitter".to_string(),
            )
            .await
            .unwrap();
            assert!(res.as_object_mut().unwrap().contains_key("oauth_token"));
        }
    }
}
