use crate::{
    api_model::{
        CreateItem, Oauth2AccessTokenRequest, Oauth2AccessTokenResponse, Oauth2AuthUrlRequest,
        Oauth2AuthorizeTokenRequest, Platforms, Search, SortOrder,
    },
    command_line_interface::CliOptions,
    db_model::{
        Oauth2Flow, ACCESS_TOKEN, OAUTH2_FLOW, PLATFORM, REDIRECT_URI, REFRESH_TOKEN, TOKEN_TYPE,
    },
    error::{Error, ErrorContext, Result},
    internal_api::{self, search},
    schema::Schema,
};
use lazy_static::lazy_static;
use oauth2::{
    basic::{BasicClient, BasicTokenResponse},
    reqwest::http_client,
    AuthUrl, AuthorizationCode, ClientId, ClientSecret, CsrfToken, RedirectUrl, RefreshToken,
    Scope, TokenResponse, TokenUrl,
};
use reqwest::{header::CONTENT_TYPE, StatusCode};
use rusqlite::Transaction;
use serde::Deserialize;
use serde_json::json;
use std::{collections::HashMap, env};
use tokio::task;
use tracing::debug;

pub fn auth_url(payload: Oauth2AuthUrlRequest) -> Result<String> {
    debug!("Creating auth url for {:?}", payload.platform);

    let client = get_client(&payload.platform, payload.redirect_uri)?;

    let (authorize_url, _csrf_state) = client
        .authorize_url(CsrfToken::new_random)
        .add_scopes(payload.scopes.into_iter().map(Scope::new))
        .url();

    Ok(authorize_url.to_string())
}

pub fn access_token(
    tx: &Transaction,
    schema: &mut Schema,
    payload: Oauth2AccessTokenRequest,
) -> Result<Oauth2AccessTokenResponse> {
    debug!("Retrieving access token for {:?}", payload.platform);

    let oauth2_item = get_oauth2_flow(tx, schema, &payload.platform)?;

    if token_should_be_refreshed(oauth2_item.access_token.clone(), &payload.platform)? {
        debug!("Refreshing token...");

        let token = exchange_refresh_token(&oauth2_item)?;

        internal_api::update_item_tx(
            tx,
            schema,
            &oauth2_item.base.id,
            json!({
                ACCESS_TOKEN: token.access_token().secret(),
                REFRESH_TOKEN: token.refresh_token().expect("missing refresh token").secret(),
                TOKEN_TYPE: token.token_type().as_ref().to_string(),
            })
            .as_object()
            .expect("Invalid json object for oauth2")
            .clone()
            .into_iter()
            .collect(),
        )?;

        Ok(Oauth2AccessTokenResponse {
            access_token: token.access_token().secret().clone(),
            token_type: token.token_type().as_ref().to_string(),
        })
    } else {
        Ok(Oauth2AccessTokenResponse {
            access_token: oauth2_item.access_token,
            token_type: oauth2_item.token_type,
        })
    }
}

pub fn authorize(
    tx: &Transaction,
    schema: &mut Schema,
    owner: &str,
    cli: &CliOptions,
    payload: Oauth2AuthorizeTokenRequest,
) -> Result<Oauth2AccessTokenResponse> {
    debug!("Authorizing token for {:?}", payload.platform);

    let token = exchange_code(&payload)?;

    // It's allowed to call authorize many times for the same platform,
    // re-use the Oauth2Flow item in the database if exist.
    match get_oauth2_flow(tx, schema, &payload.platform) {
        Ok(item) => {
            // Update already existing item in the database
            internal_api::update_item_tx(
                tx,
                schema,
                &item.base.id,
                json!({
                    ACCESS_TOKEN: token.access_token().secret(),
                    REFRESH_TOKEN: token.refresh_token().expect("missing refresh token").secret(),
                    TOKEN_TYPE: token.token_type().as_ref().to_string(),
                    REDIRECT_URI: payload.redirect_uri,
                })
                .as_object()
                .expect("Invalid json object for oauth2")
                .clone()
                .into_iter()
                .collect(),
            )?;

            debug!(
                "Access token for {:?}, updated {}",
                payload.platform, item.base.id
            );
        }
        Err(_) => {
            // Create new Oauth2Flow item
            let item = CreateItem {
                _type: OAUTH2_FLOW.to_string(),
                fields: json!({
                    ACCESS_TOKEN: token.access_token().secret(),
                    REFRESH_TOKEN: token.refresh_token().expect("missing refresh token").secret(),
                    TOKEN_TYPE: token.token_type().as_ref().to_string(),
                    REDIRECT_URI: payload.redirect_uri,
                    PLATFORM: payload.platform,
                })
                .as_object()
                .expect("Invalid json object for oauth2")
                .clone()
                .into_iter()
                .collect(),
                ..Default::default()
            };

            let id = internal_api::create_item_tx(tx, schema, item, owner, cli)?;
            debug!("Access token for {:?}, created {id}", payload.platform);
        }
    };

    Ok(Oauth2AccessTokenResponse {
        access_token: token.access_token().secret().clone(),
        token_type: token.token_type().as_ref().to_string(),
    })
}

struct PlatformConfiguration {
    client_id_env: &'static str,
    client_secret_env: &'static str,
    auth_url: AuthUrl,
    token_url: TokenUrl,
    token_info_url: String,
}

lazy_static! {
    static ref PLATFORM_CONFIG: HashMap<Platforms, PlatformConfiguration> = HashMap::from([(
        Platforms::Gitlab,
        PlatformConfiguration {
            client_id_env: "GITLAB_CLIENT_ID",
            client_secret_env: "GITLAB_CLIENT_SECRET",
            auth_url: AuthUrl::new("https://gitlab.memri.io/oauth/authorize".to_string()).unwrap(),
            token_url: TokenUrl::new("https://gitlab.memri.io/oauth/token".to_string()).unwrap(),
            token_info_url: "https://gitlab.memri.io/oauth/token/info".to_string()
        },
    )]);
}

fn exchange_refresh_token(oauth2_item: &Oauth2Flow) -> Result<BasicTokenResponse> {
    let client = get_client(&oauth2_item.platform, oauth2_item.redirect_uri.clone())?;

    let res = task::block_in_place(|| {
        client
            .exchange_refresh_token(&RefreshToken::new(oauth2_item.refresh_token.clone()))
            .request(http_client)
    })?;
    Ok(res)
}

fn exchange_code(payload: &Oauth2AuthorizeTokenRequest) -> Result<BasicTokenResponse> {
    let client = get_client(&payload.platform, payload.redirect_uri.clone())?;

    let res = task::block_in_place(|| {
        client
            .exchange_code(AuthorizationCode::new(payload.auth_code.clone()))
            .request(http_client)
    })?;

    Ok(res)
}

/// Creates oauth2 client for specific platform
fn get_client(platform: &Platforms, redirect_uri: String) -> Result<BasicClient> {
    let config = PLATFORM_CONFIG.get(platform).ok_or(Error {
        code: StatusCode::INTERNAL_SERVER_ERROR,
        msg: format!("Missing configuration for {platform:?}"),
    })?;

    let client_id = ClientId::new(
        env::var(config.client_id_env).context(|| format!("missing: {}", config.client_id_env))?,
    );
    let client_secret = ClientSecret::new(
        env::var(config.client_secret_env)
            .context(|| format!("missing: {}", config.client_secret_env))?,
    );

    Ok(BasicClient::new(
        client_id,
        Some(client_secret),
        config.auth_url.clone(),
        Some(config.token_url.clone()),
    )
    .set_redirect_uri(RedirectUrl::new(redirect_uri)?))
}

fn get_oauth2_flow(tx: &Transaction, schema: &Schema, platform: &Platforms) -> Result<Oauth2Flow> {
    let query = Search {
        _type: Some(OAUTH2_FLOW.to_string()),
        deleted: Some(false),
        limit: 5,
        sort_order: SortOrder::Desc,
        other_properties: HashMap::from([(
            PLATFORM.to_string(),
            serde_json::to_value(&platform).expect("Cannot serialize Platforms to Value"),
        )]),
        ..Default::default()
    };

    let mut items = search(tx, schema, query)?;

    debug_assert!(
        items.len() <= 1,
        "There is more than one Oauth2Flow item in the DB for platform {platform:?}, {items:#?}"
    );

    let oauth2_item = items.pop().ok_or(Error {
        code: StatusCode::BAD_REQUEST,
        msg: format!("Could not find Oauth2Flow for {platform:?}",),
    })?;

    let oauth2_item: Oauth2Flow = serde_json::from_value(oauth2_item)
        .context(|| format!("Cannot deserialize Oauth2Flow for platform {platform:?}",))?;

    debug!("Found Oauth2 item {}", oauth2_item.base.id);

    Ok(oauth2_item)
}

#[derive(Debug, Deserialize)]
struct GitlabTokenInfo {
    expires_in: u32,
    // resource_owner_id: usize,
    // scope: Vec<String>,
    // application: Value,
    // created_at: u32,
}
const EXPIRY_THRESHOLD: u32 = 60;

fn token_should_be_refreshed(access_token: String, platform: &Platforms) -> Result<bool> {
    let token_info_url = &PLATFORM_CONFIG
        .get(platform)
        .ok_or(Error {
            code: StatusCode::INTERNAL_SERVER_ERROR,
            msg: format!("Missing configuration for {platform:?}"),
        })?
        .token_info_url;

    task::block_in_place(move || -> Result<bool> {
        let client = reqwest::blocking::Client::new();
        let res = client
            .get(token_info_url)
            .header(CONTENT_TYPE, "application/json")
            .bearer_auth(access_token)
            .send()
            .context(|| format!("Failed to retrieve token information from {token_info_url}",))?;

        if res.status() != StatusCode::OK {
            debug!(
                "Error GET {token_info_url} response {:#?}",
                res.json::<serde_json::Value>()
            );
            Ok(true)
        } else {
            let res: GitlabTokenInfo = res.json()?;
            debug!("Ok GET {token_info_url} response {res:#?}");

            // Token is going to expiry very soon
            Ok(res.expires_in < EXPIRY_THRESHOLD)
        }
    })
}
