use crate::{
    api_model::CreateItem,
    command_line_interface::CliOptions,
    database_api,
    db_model::{PluginRunItem, CONTAINER_ID, CONTAINER_IMAGE},
    error::{Error, Result},
    internal_api,
    internal_api::{new_random_string, search},
    plugin_auth_crypto::DatabaseKey,
    schema::Schema,
};
use rusqlite::Transaction;
use serde_json::{json, Value};
use std::{collections::HashMap, process::Command, sync::atomic::AtomicU32};
use tracing::{debug, info, warn};
use warp::http::status::StatusCode;

/// Run a plugin, making sure that the correct ENV variables and settings are passed
/// to the containerization / deployment processes.
///
/// Internally passes to docker / kubernetes / scripts
/// depending on how Pod is configured by the user.
#[allow(clippy::too_many_arguments)]
pub fn run_plugin_container(
    tx: &Transaction,
    schema: &Schema,
    triggered_by_item_id: &str,
    plugin: &PluginRunItem,
    pod_owner: &str,
    database_key: &DatabaseKey,
    cli_options: &CliOptions,
) -> Result<()> {
    info!(
        "Trying to run plugin container for target_item_id {}",
        plugin.target_item_id
    );
    let target_item = internal_api::get_item_tx(tx, schema, plugin.target_item_id.as_str())?;
    let target_item = target_item.into_iter().next().ok_or_else(|| Error {
        code: StatusCode::BAD_REQUEST,
        msg: format!(
            "Failed to find target item {} to run a plugin against",
            plugin.target_item_id
        ),
    })?;

    // TODO: that should be not passed to the plugin, plugin should request that data from the DB
    // passing plugin run id is enough.
    let target_item_json = serde_json::to_string(&target_item)?;
    let auth = database_key.create_plugin_auth()?;
    let auth = serde_json::to_string(&auth)?;

    let script_override = cli_options
        .insecure_plugin_script
        .iter()
        .find(|(image, _script)| image == &plugin.container_image)
        .map(|(_image, script)| script);

    if let Some(script_path) = script_override {
        run_local_script(
            &plugin.container_image,
            script_path,
            &target_item_json,
            pod_owner,
            &auth,
            triggered_by_item_id,
            cli_options,
        )
    } else if cli_options.use_kubernetes {
        run_kubernetes_container(
            &target_item_json,
            pod_owner,
            &auth,
            triggered_by_item_id,
            plugin,
            cli_options,
        )
    } else {
        run_docker_container(
            &target_item_json,
            pod_owner,
            &auth,
            triggered_by_item_id,
            plugin,
            cli_options,
        )
    }
}

fn run_local_script(
    _container: &str,
    plugin_path: &str,
    target_item: &str,
    pod_owner: &str,
    pod_auth: &str,
    triggered_by_item_id: &str,
    cli_options: &CliOptions,
) -> Result<()> {
    let pod_full_address = callback_address(cli_options, false);

    let args: Vec<String> = Vec::new();
    let mut env_vars = HashMap::new();
    env_vars.insert("POD_FULL_ADDRESS", pod_full_address.as_str());
    env_vars.insert("POD_TARGET_ITEM", target_item);
    env_vars.insert("POD_PLUGINRUN_ID", triggered_by_item_id);
    env_vars.insert("POD_OWNER", pod_owner);
    env_vars.insert("POD_AUTH_JSON", pod_auth);

    run_any_command(plugin_path, &args, &env_vars, triggered_by_item_id)
}

// Example:
// docker run \
//     --network=host \
//     --env=POD_FULL_ADDRESS="http://localhost:3030" \
//     --env=POD_TARGET_ITEM="{...json...}" \
//     --env=POD_OWNER="...64-hex-chars..." \
//     --env=POD_AUTH_JSON="{...json...}" \
//
//     --name="$containerImage-$trigger_item_id" \
//     --rm \
//     -- \
//     "$containerImage"
fn run_docker_container(
    target_item_json: &str,
    pod_owner: &str,
    pod_auth: &str,
    triggered_by_item_id: &str,
    plugin: &PluginRunItem,
    cli_options: &CliOptions,
) -> Result<()> {
    let docker_network = match &cli_options.plugins_docker_network {
        Some(net) => net.to_string(),
        None => "host".to_string(),
    };

    let args: Vec<String> = vec![
        "run".to_string(),
        "--rm".to_string(),
        format!("--network={}", docker_network),
        format!("-h={triggered_by_item_id}"),
        format!(
            "--env=POD_FULL_ADDRESS={}",
            callback_address(cli_options, true)
        ),
        format!("--env=POD_TARGET_ITEM={}", target_item_json),
        format!("--env=POD_PLUGINRUN_ID={}", triggered_by_item_id),
        format!("--env=POD_OWNER={}", pod_owner),
        format!("--env=POD_AUTH_JSON={}", pod_auth),
        // Makes logs from the Python visible on the POD console immediately
        "--env=PYTHONUNBUFFERED=1".to_string(),
        format!(
            "--name={}",
            sanitize_docker_name(plugin.container_id.as_str())
        ),
        "--".to_string(),
        plugin.container_image.clone(),
    ];
    let envs: HashMap<&str, &str> = HashMap::new();
    run_any_command("docker", &args, &envs, triggered_by_item_id)
}

/// Example:
/// kubectl run $owner-$containerImage-$targetItem-$randomHex
///     --image="$containerImage" \
///     --env=POD_FULL_ADDRESS="http://localhost:3030" \
///     --env=POD_TARGET_ITEM="{...json...}" \
///     --env=POD_OWNER="...64-hex-chars..." \
///     --env=POD_AUTH_JSON="{...json...}" \
fn run_kubernetes_container(
    target_item_json: &str,
    pod_owner: &str,
    pod_auth: &str,
    triggered_by_item_id: &str,
    plugin: &PluginRunItem,
    cli_options: &CliOptions,
) -> Result<()> {
    let args: Vec<String> = vec![
        "run".to_string(),
        "--restart=Never".to_string(),
        plugin.container_id.clone(),
        format!("--labels=app={},type=plugin", plugin.container_id),
        "--port=8080".to_string(),
        "--image-pull-policy=Always".to_string(),
        format!("--image={}", plugin.container_image),
        format!(
            "--env=POD_FULL_ADDRESS={}",
            callback_address(cli_options, false)
        ),
        format!("--env=POD_TARGET_ITEM={}", target_item_json),
        format!("--env=POD_PLUGINRUN_ID={}", triggered_by_item_id),
        format!("--env=POD_OWNER={}", pod_owner),
        format!("--env=POD_AUTH_JSON={}", pod_auth),
        format!("--env=PLUGIN_DNS={}", plugin.webserver_url),
        "--env=PYTHONUNBUFFERED=1".to_string(),
    ];
    run_any_command("kubectl", &args, &HashMap::new(), triggered_by_item_id)
}

pub fn get_logs(tx: &Transaction, plugin_run_id: &str, cli_options: &CliOptions) -> Result<Value> {
    let schema = database_api::get_schema(tx).unwrap();

    let json = json!({
        "id": plugin_run_id,
    });
    let query = serde_json::from_value(json).unwrap();
    let result = search(tx, &schema, query).unwrap();
    let result = result.into_iter().next().unwrap();
    let container_id = result.get("containerId").unwrap().as_str().unwrap();
    let args: Vec<String> = vec!["logs".to_string(), container_id.to_string()];

    if cli_options.use_kubernetes {
        //kubernetes logs
        let cmd = "kubectl".to_string();
        let logs = run_any_command_for_logs(cmd, args);
        Ok(json!({
            "logs": logs,
        }))
    } else {
        //docker logs
        let cmd = "docker".to_string();
        let logs = run_any_command_for_logs(cmd, args);
        Ok(json!({
            "logs": logs,
        }))
    }
}

fn run_any_command_for_logs(cmd: String, args: Vec<String>) -> String {
    let debug_args = args
        .iter()
        .map(|p| escape_bash_arg(p))
        .collect::<Vec<_>>()
        .join(" ");
    info!("Getting logs {} {}", cmd, debug_args);

    let output = Command::new(cmd)
        .args(args)
        .output()
        .expect("failed to fetch logs");
    String::from_utf8_lossy(&output.stdout).to_string()
}

fn run_any_command(
    cmd: &str,
    args: &[String],
    envs: &HashMap<&str, &str>,
    container_id: &str,
) -> Result<()> {
    let debug_envs = envs
        .iter()
        .map(|(a, b)| format!("{}={} ", escape_bash_arg(a), escape_bash_arg(b)))
        .collect::<Vec<_>>()
        .join("");
    let debug_args = args
        .iter()
        .map(|p| escape_bash_arg(p))
        .collect::<Vec<_>>()
        .join(" ");
    info!("Starting command {}{} {}", debug_envs, cmd, debug_args);

    let command = Command::new(cmd).args(args).envs(envs).spawn();
    match command {
        Ok(_child) => {
            debug!(
                "Successfully started {} process for Plugin container item {}",
                cmd, container_id
            );
            Ok(())
        }
        Err(err) => Err(Error {
            code: StatusCode::INTERNAL_SERVER_ERROR,
            msg: format!(
                "Failed to execute {}, a plugin container triggered by item.rowid {}, {}",
                cmd, container_id, err
            ),
        }),
    }
}

/// Determine the callback address to use for plugins.
/// A callback address is an environment variable passed to the Plugin
/// so that the plugin can call back Pod when it needs to.
fn callback_address(cli_options: &CliOptions, docker_magic: bool) -> String {
    if let Some(address_override) = &cli_options.plugins_callback_address {
        address_override.to_string()
    } else {
        // The plugin container needs to have access to the host
        // This is currently done differently on MacOS and Linux
        // https://stackoverflow.com/questions/24319662/from-inside-of-a-docker-container-how-do-i-connect-to-the-localhost-of-the-mach
        let pod_domain = if docker_magic && !cfg!(target_os = "linux") {
            "host.docker.internal"
        } else {
            "localhost"
        };
        let is_https = cli_options.insecure_non_tls.is_none() && !cli_options.non_tls;
        let schema = if is_https { "https" } else { "http" };
        format!("{}://{}:{}", schema, pod_domain, cli_options.port)
    }
}

pub fn sanitize_docker_name(input: &str) -> String {
    input
        .chars()
        .map(|c| {
            if c.is_ascii_alphanumeric() || "-_".contains(c) {
                c
            } else {
                '_'
            }
        })
        .collect()
}

/// Debug-print a bash argument.
/// Never use this for running real code, but for debugging that's good enough.
///
/// From bash manual:
/// > Enclosing  characters in single quotes preserves the literal value of each character
/// > within the quotes. A single quote may not occur between single quotes,
/// > even when preceded by a backslash.
pub fn escape_bash_arg(str: &str) -> String {
    let ok = str
        .chars()
        .all(|c| c.is_ascii_alphanumeric() || "_-+=%".contains(c));
    if ok {
        str.to_string()
    } else {
        let quoted = str.replace('\'', "'\\''"); // end quoting, append the literal, start quoting
        format!("'{}'", quoted)
    }
}

/// Get next available port number for PluginRun instance
pub fn get_free_port(cli: &CliOptions) -> u32 {
    if (cli.use_kubernetes && cli.plugins_public_domain.is_some())
        || cli.plugins_docker_network.is_some()
    {
        // Container have unique IPs, same port might be used
        return 8080;
    }

    // Container use host network stack, give unique port number
    static FREE_PORT_NUMBER: AtomicU32 = AtomicU32::new(8008);
    FREE_PORT_NUMBER.fetch_add(1, std::sync::atomic::Ordering::Relaxed)
}

/// Figure out correct URL depending on the environnement the Plugin will be run on.
pub fn resolve_webserver_url(item: &CreateItem, cli: &CliOptions) -> Result<String> {
    let plugin_run_id = item.id.as_ref().ok_or(Error {
        code: StatusCode::BAD_REQUEST,
        msg: "Could not get required field Id".to_string(),
    })?;

    let container_id = item
        .fields
        .get(CONTAINER_ID)
        .and_then(|val| val.as_str())
        .ok_or(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("Could not get required field {CONTAINER_ID}"),
        })?;

    let webserver_url = if cli.use_kubernetes {
        // Kubernetes
        if let Some(base) = &cli.plugins_public_domain {
            format!("https://{}.{}", container_id, base)
        } else {
            warn!("Plugin public DNS is not configured for the Pod but a kubernetes Plugin is run. Using localhost as the public domain...");
            "http://localhost".to_string()
        }
    } else {
        // Docker
        match &cli.plugins_docker_network {
            // There is a custom network provided, access the container from the POD using it's hostname,
            // Which is the ID of PluginRun object in the Database.
            // Note: it will work out of the box, if the POD is also running in the container,
            // If POD runs on the host, additional DNS service must run to resolve containers hostnames to IPs
            // The DNS can be executed by calling examples/run_docker_dns_service.sh
            // TODO: use docker API to start such a service? That will change host network configuration.
            Some(_net) => format!("http://{plugin_run_id}"),
            // Network is not set, use host mode, and access the container from the POD using localhost
            None => "http://localhost".to_string(),
        }
    };

    Ok(webserver_url)
}

pub fn get_container_id(item: &CreateItem, cli: &CliOptions, pod_owner: &str) -> Result<String> {
    let container_image = item
        .fields
        .get(CONTAINER_IMAGE)
        .and_then(|val| val.as_str())
        .ok_or(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("Could not get required field {CONTAINER_IMAGE}"),
        })?;

    let plugin_run_id = item.id.as_ref().ok_or(Error {
        code: StatusCode::BAD_REQUEST,
        msg: "Could not get required field Id".to_string(),
    })?;

    let container_id = if cli.use_kubernetes {
        format!(
            "c{}-{}-{}-{}",
            pod_owner
                .chars()
                .filter(|c| c.is_ascii_alphanumeric())
                .take(10)
                .collect::<String>(),
            container_image
                .chars()
                .filter(|c| c.is_ascii_alphanumeric())
                .take(20)
                .collect::<String>(),
            plugin_run_id
                .chars()
                .filter(|c| c.is_ascii_alphanumeric())
                .take(20)
                .collect::<String>(),
            new_random_string(8)
        )
    } else {
        format!(
            "{}-{}-{}",
            pod_owner
                .chars()
                .filter(|c| c.is_ascii_alphanumeric())
                .collect::<String>(),
            container_image
                .chars()
                .filter(|c| c.is_ascii_alphanumeric())
                .collect::<String>(),
            plugin_run_id
                .chars()
                .filter(|c| c.is_ascii_alphanumeric())
                .collect::<String>()
        )
    };

    Ok(container_id)
}
