use crate::{
    api_model::{Search, SortOrder},
    command_line_interface::CliOptions,
    db_model::{ItemBase, PluginRunItem, Trigger},
    error::{Error, ErrorContext, Result},
    internal_api::search,
    json_utils::get_property_value,
    schema::Schema,
    triggers::AfterAction,
};
use reqwest::header::CONTENT_TYPE;
use rusqlite::Transaction;
use serde_json::{json, Value};
use tokio::task;
use tracing::{debug, info, trace};
use warp::http::status::StatusCode;

pub fn on_new_item(
    tx: &Transaction,
    schema: &Schema,
    item_base: &ItemBase,
    item_created: &Value,
    cli: &CliOptions,
) -> Result<AfterAction> {
    let query = Search {
        _type: Some("Trigger".to_string()),
        deleted: Some(false),
        limit: u64::MAX,
        // Return triggers in order they were inserted
        sort_order: SortOrder::Asc,
        ..Default::default()
    };

    let triggers = search(tx, schema, query)?
        .into_iter()
        .map(serde_json::from_value)
        .collect::<std::result::Result<Vec<Trigger>, serde_json::Error>>()
        .context_str("Failed to retrieve Triggers from the DB")?;

    let mut after_action = AfterAction::default();

    for trigger in triggers {
        if trigger.trigger_on == item_base._type {
            let created_after =
                get_property_value(item_created, &trigger.filter_created_after_property_name)?
                    .as_i64()
                    .ok_or(Error {
                        code: StatusCode::BAD_REQUEST,
                        msg: "Cannot parse property to i64".to_string(),
                    })?;

            if created_after > trigger.filter_created_after {
                let plugin_instance = get_plugin_instance(tx, schema, &trigger.plugin_run_id);

                let mut delete_trigger = false;

                match plugin_instance {
                    Ok(plugin_instance) => {
                        if let Some(status) = &plugin_instance.status {
                            match status.as_str() {
                                "done" | "error" => {
                                    // Plugin was registered for the trigger that has finished, delete the trigger
                                    info!(
                                        "Deleting trigger {}, because plugin {} has status {status}",
                                        trigger.id, trigger.plugin_run_id
                                    );
                                    delete_trigger = true;
                                }
                                // Plugin still works, notify it
                                _ => notify_plugin(&trigger, item_base, &plugin_instance, cli)?,
                            }
                        } else {
                            trace!(
                                "There is no status field for plugin {}, yet",
                                trigger.plugin_run_id
                            );
                        }
                    }
                    Err(e) => {
                        info!("Deleting trigger {}, because of {e}", trigger.id);
                        delete_trigger = true;
                    }
                }

                if delete_trigger {
                    after_action.items_to_delete.insert(trigger.id);
                }
            }
        }
    }

    Ok(after_action)
}

fn get_plugin_instance(
    tx: &Transaction,
    schema: &Schema,
    plugin_run_id: &str,
) -> Result<PluginRunItem> {
    let query = Search {
        id: Some(plugin_run_id.to_string()),
        deleted: Some(false),
        limit: 1,
        ..Default::default()
    };

    let plugin_instance = search(tx, schema, query)?.pop().ok_or(Error {
        code: StatusCode::BAD_REQUEST,
        msg: format!(
            "Could not find PluginRun {}, maybe it's deleted?",
            plugin_run_id
        ),
    })?;

    let plugin_instance: PluginRunItem = serde_json::from_value(plugin_instance)
        .context(|| format!("Cannot parse PluginRun with id {}", plugin_run_id))?;

    Ok(plugin_instance)
}

fn notify_plugin(
    trigger: &Trigger,
    item_base: &ItemBase,
    plugin_instance: &PluginRunItem,
    cli: &CliOptions,
) -> Result<()> {
    let host = &plugin_instance.webserver_url;

    let endpoint = if cli.use_kubernetes && cli.plugins_public_domain.is_some() {
        // On kubernetes reach the plugin via 80/443 port
        format!("{host}/{}", trigger.action)
    } else {
        let port = plugin_instance.webserver_port;
        format!("{host}:{port}/{}", trigger.action)
    };

    let body =
        json!({ "item_id": item_base.id.clone(), "trigger_id": trigger.id.clone() }).to_string();

    debug!(
        "Going to trigger PluginRun instance {} by Trigger {} because of arrival of {}",
        trigger.plugin_run_id, trigger.id, item_base._type
    );
    debug!("POST: {endpoint}, with body {body:#}");

    task::block_in_place(move || {
        let client = reqwest::blocking::Client::new();
        let res = client
            .post(&endpoint)
            .header(CONTENT_TYPE, "application/json")
            .body(body.clone())
            .send()
            .context(|| {
                format!(
                    "Failed to trigger plugin instance {}, endpoint {endpoint}",
                    trigger.plugin_run_id
                )
            })?;

        debug!("Got POST {} response {res:?}", trigger.action);

        if res.status() != StatusCode::OK {
            return Err(Error {
                                code: res.status(),
                                msg: format!("Failure while triggering the plugin. PluginRun = {}, action = {} body = {}, error = {:?}",
                                    trigger.plugin_run_id, trigger.action, body, res.text())
                                });
        }

        Ok(())
    })
}
