use crate::error::{Error, Result};
use lazy_static::lazy_static;
use regex::Regex;
use serde::{Deserialize, Serialize};
use std::{
    collections::{HashMap, HashSet},
    convert::TryInto,
};
use time::OffsetDateTime;
use warp::http::StatusCode;

/// See `README.md#understanding-the-schema` to understand possible
/// property types and their meaning
#[derive(Serialize, Deserialize, Debug, Copy, Clone, PartialEq, Eq)]
pub enum SchemaPropertyType {
    Text,
    Integer,
    Real,
    Bool,
    DateTime,
}

pub const BASE_PROPERTIES: &[&str] = &[
    "id",
    "type",
    "dateCreated",
    "dateModified",
    "dateServerModified",
    "deleted",
];

impl std::fmt::Display for SchemaPropertyType {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl SchemaPropertyType {
    pub fn from_string(str: &str) -> std::result::Result<SchemaPropertyType, String> {
        match str {
            "Text" => Ok(SchemaPropertyType::Text),
            "Integer" => Ok(SchemaPropertyType::Integer),
            "Real" => Ok(SchemaPropertyType::Real),
            "Bool" => Ok(SchemaPropertyType::Bool),
            "DateTime" => Ok(SchemaPropertyType::DateTime),
            _ => Err(format!(
                "Failed to parse {} into {}",
                str,
                std::any::type_name::<SchemaPropertyType>()
            )),
        }
    }
}

/// Name of type defining schema
pub const ITEM_PROPERTY_SCHEMA: &str = "ItemPropertySchema";
pub const ITEM_EDGE_SCHEMA: &str = "ItemEdgeSchema";

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct SchemaItem {
    pub item_type: String,
    pub property_name: String,
    pub value_type: SchemaPropertyType,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct SchemaEdge {
    pub edge_name: String,
    pub source_type: String,
    pub target_type: String,
}

#[derive(Debug)]
pub struct Schema {
    /// United properties across all types, for example:
    ///     "age" -> Integer
    property_types: HashMap<String, SchemaPropertyType>,
    /// All types and their properties, for example:
    ///     "Person", "age" -> Integer
    type_properties: HashMap<String, HashMap<String, SchemaPropertyType>>,
    /// All types and their edges, for example:
    ///     "Person", "friend" -> "Person"
    type_edges: HashMap<String, HashMap<String, Vec<String>>>,
}

impl Schema {
    pub fn empty() -> Schema {
        Schema {
            property_types: HashMap::new(),
            type_properties: HashMap::new(),
            type_edges: HashMap::new(),
        }
    }
    pub fn set_type_unchecked(
        &mut self,
        item_type: String,
        property_name: String,
        value_type: SchemaPropertyType,
    ) {
        self.property_types
            .insert(property_name.to_string(), value_type);
        self.type_properties
            .entry(item_type)
            .or_insert_with(HashMap::new)
            .insert(property_name, value_type);
    }

    pub fn set_edgetype_unchecked(
        &mut self,
        source_type: String,
        edge_name: String,
        target_type: String,
    ) {
        let target_types = self
            .type_edges
            .entry(source_type)
            .or_insert_with(HashMap::new)
            .entry(edge_name)
            .or_insert_with(Vec::new);

        if !target_types.contains(&target_type) {
            target_types.push(target_type);
        }
    }
    pub fn property(&self, property_name: &str) -> Option<&SchemaPropertyType> {
        self.property_types.get(property_name)
    }
    pub fn item_property(
        &self,
        item_type: &str,
        property_name: &str,
    ) -> Option<&SchemaPropertyType> {
        self.type_properties
            .get(item_type)
            .and_then(|i| i.get(property_name))
    }
    /// The total number of properties united for all types.
    /// (E.g. "Person.age" and "Computer.age" only count once, because that's one property "age".)
    pub fn all_properties_len(&self) -> usize {
        self.property_types.len()
    }

    pub fn item_edge(&self, item_type: &str, edge_name: &str) -> Option<Vec<String>> {
        self.type_edges.get(item_type)?.get(edge_name).cloned()
    }

    pub fn is_base_property(&self, property_name: &str) -> bool {
        BASE_PROPERTIES.contains(&property_name)
    }
}

/// Validation of _new_ item ids. Note that it is not applied to already existing
/// ids or endpoints that access already existing ids.
pub fn validate_create_item_id(item_id: &str) -> Result<()> {
    lazy_static! {
        static ref REGEXP: Regex =
            Regex::new(r"^[a-zA-Z0-9_-]{6,36}$").expect("Cannot create regex");
    }
    if !REGEXP.is_match(item_id) {
        Err(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!(
                "Item id '{}' does not satisfy the format {} (use 32 random hex characters if in doubt on item id creation)",
                item_id,
                REGEXP.as_str()
            ),
        })
    } else {
        Ok(())
    }
}

fn validate_property_name_syntax(property: &str) -> Result<()> {
    lazy_static! {
        static ref REGEXP: Regex =
            Regex::new(r"^[a-zA-Z][_a-zA-Z0-9]{0,30}[a-zA-Z0-9]$").expect("Cannot create regex");
    }
    if !REGEXP.is_match(property) {
        Err(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!(
                "Property name {} does not satisfy the format {}",
                property,
                REGEXP.as_str()
            ),
        })
    } else {
        Ok(())
    }
}

/// All item properties should be of this format
pub fn validate_property_name(property: &str) -> Result<()> {
    validate_property_name_syntax(property)?;

    if BLOCKLIST_COLUMN_NAMES.contains(&property.to_lowercase()) {
        Err(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("Blocklisted item property {}", property),
        })
    } else {
        Ok(())
    }
}

/// Returns the DateTime in Memri format,
/// See `README.md#understanding-the-schema` for the exact definition.
/// (In short, it's the EPOCH milliseconds.)
pub fn utc_millis() -> i64 {
    (OffsetDateTime::now_utc().unix_timestamp_nanos() / 1_000_000)
        .try_into()
        .expect("Current time is outside of Memri DateTime range (unsigned 64-bit integer)")
}

lazy_static! {
    static ref BLOCKLIST_COLUMN_NAMES: HashSet<String> = {
        SQLITE_RESERVED_KEYWORDS
            .iter()
            .chain(POD_ITEM_MANDATORY_PROPERTIES)
            .map(|w| w.to_lowercase())
            .collect()
    };
}

const POD_ITEM_MANDATORY_PROPERTIES: &[&str] = &[
    "rowid",
    "id",
    "type",
    "dateCreated",
    "dateModified",
    "dateServerModified",
    "deleted",
    "source",
    "target",
    "edges",
    "allEdges",
    "forwardEdges",
    "reverseEdges",
    "backwardEdges",
];

/// SQLite keywords taken from https://www.sqlite.org/lang_keywords.html
const SQLITE_RESERVED_KEYWORDS: &[&str] = &[
    "ABORT",
    "ACTION",
    "ADD",
    "AFTER",
    "ALL",
    "ALTER",
    "ALWAYS",
    "ANALYZE",
    "AND",
    "AS",
    "ASC",
    "ATTACH",
    "AUTOINCREMENT",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BY",
    "CASCADE",
    "CASE",
    "CAST",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "COMMIT",
    "CONFLICT",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "DATABASE",
    "DEFAULT",
    "DEFERRABLE",
    "DEFERRED",
    "DELETE",
    "DESC",
    "DETACH",
    "DISTINCT",
    "DO",
    "DROP",
    "EACH",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXCLUDE",
    "EXCLUSIVE",
    "EXISTS",
    "EXPLAIN",
    "FAIL",
    "FILTER",
    "FIRST",
    "FOLLOWING",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "GENERATED",
    "GLOB",
    "GROUP",
    "GROUPS",
    "HAVING",
    "IF",
    "IGNORE",
    "IMMEDIATE",
    "IN",
    "INDEX",
    "INDEXED",
    "INITIALLY",
    "INNER",
    "INSERT",
    "INSTEAD",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "JOIN",
    "KEY",
    "LAST",
    "LEFT",
    "LIKE",
    "LIMIT",
    "MATCH",
    "NATURAL",
    "NO",
    "NOT",
    "NOTHING",
    "NOTNULL",
    "NULL",
    "NULLS",
    "OF",
    "OFFSET",
    "ON",
    "OR",
    "ORDER",
    "OTHERS",
    "OUTER",
    "OVER",
    "PARTITION",
    "PLAN",
    "PRAGMA",
    "PRECEDING",
    "PRIMARY",
    "QUERY",
    "RAISE",
    "RANGE",
    "RECURSIVE",
    "REFERENCES",
    "REGEXP",
    "REINDEX",
    "RELEASE",
    "RENAME",
    "REPLACE",
    "RESTRICT",
    "RIGHT",
    "ROLLBACK",
    "ROW",
    "ROWS",
    "SAVEPOINT",
    "SELECT",
    "SET",
    "TABLE",
    "TEMP",
    "TEMPORARY",
    "THEN",
    "TIES",
    "TO",
    "TRANSACTION",
    "TRIGGER",
    "UNBOUNDED",
    "UNION",
    "UNIQUE",
    "UPDATE",
    "USING",
    "VACUUM",
    "VALUES",
    "VIEW",
    "VIRTUAL",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
    "WITHOUT",
];
