//
// Pod triggers that should "run" on particular DB/data changes (e.g. item insertion)
//

use std::collections::HashSet;

use crate::{
    api_model::CreateItem,
    command_line_interface::CliOptions,
    database_api::{self, in_read_transaction, in_write_transaction},
    db_connection_manager::DBConnection,
    db_model::{ItemBase, PluginRunItem, CONTAINER_ID, WEBSERVER_PORT, WEBSERVER_URL},
    error::{Error, ErrorContext, Result},
    internal_api::{self, get_item_tx},
    plugin_auth_crypto::DatabaseKey,
    plugin_run, plugin_trigger,
    schema::{Schema, SchemaEdge, SchemaItem, ITEM_EDGE_SCHEMA, ITEM_PROPERTY_SCHEMA},
    triggers::SchemaAdditionChange::*,
};

use rusqlite::Transaction as Tx;
use serde_json::json;
use tracing::error;
use warp::http::StatusCode;

#[derive(Debug, Eq, PartialEq)]
pub enum SchemaAdditionChange {
    NotASchema,
    NewSchemaAdded,
    OldSchema { old_id: String },
}

/// If an item is a Schema, add it to the schema.
/// Fail if an incompatible schema is attempted to be inserted.
pub fn add_item_as_schema_opt(
    tx: &Tx,
    schema: &mut Schema,
    item: &CreateItem,
) -> Result<SchemaAdditionChange> {
    if item._type == ITEM_PROPERTY_SCHEMA {
        // We'll do something ugly here.
        // We'll convert the item into JSON, and then back into the desired type
        // for type check and parsing. This is easier code-wise than to do manual conversions.
        // It only triggers for specific, rarely used items. This implementation might change later.
        let json = serde_json::to_value(item)?;
        let parsed: SchemaItem = serde_json::from_value(json)
            .context(|| format!("Parsing of Schema item {:?}, {}:{}", item, file!(), line!()))?;

        // The result of the operation depends on:
        // * the "heterogeneous" property's type
        // * this particular itemType's property type
        // * whether an ID is present in the request
        match (
            schema.property(&parsed.property_name),
            schema.item_property(&parsed.item_type, &parsed.property_name),
            &item.id,
        ) {
            (Some(old_value_type), _, _) if &parsed.value_type != old_value_type => {
                Err(Error {
                    code: StatusCode::BAD_REQUEST,
                    msg: format!(
                        "Schema for property {} is already defined to valueType {}, cannot override to {}",
                        parsed.property_name,
                        old_value_type,
                        parsed.value_type,
                    )
                })
            }
            (_, None, _) | (_, _, Some(_)) => {
                schema.set_type_unchecked(parsed.item_type, parsed.property_name, parsed.value_type);
                Ok(NewSchemaAdded)
            }
            (_, Some(old_value_type), None) if &parsed.value_type == old_value_type => {
                let old_id = database_api::get_schema_item_id_for_property(
                    tx,
                    &parsed.item_type,
                    &parsed.property_name,
                )?;
                Ok(OldSchema { old_id })
            }
            (union_value_type, Some(old_value_type), None) => {
                Err(Error {
                    code: StatusCode::INTERNAL_SERVER_ERROR,
                    msg: format!(
                        "Unexpected failure, {}.{} has valueType {}. However, this Schema should not have been allowed as {} is already defined as {:?}",
                        parsed.item_type,
                        parsed.property_name,
                        old_value_type,
                        parsed.property_name,
                        union_value_type,
                    )
                })
            }
        }
    } else if item._type == ITEM_EDGE_SCHEMA {
        let json = serde_json::to_value(item)?;
        let parsed: SchemaEdge = serde_json::from_value(json).context(|| {
            format!(
                "Parsing of Schema item edge {:?}, {}:{}",
                item,
                file!(),
                line!()
            )
        })?;

        let edge_type_option = schema.item_edge(&parsed.source_type, &parsed.edge_name);

        match edge_type_option {
            // do not add if edge already exists
            Some(edge_type_vector) if edge_type_vector.contains(&parsed.target_type) => {
                let old_id = database_api::get_schema_edge_item_id(
                    tx,
                    &parsed.edge_name,
                    &parsed.source_type,
                    &parsed.target_type,
                )?;
                Ok(OldSchema { old_id })
            }
            // in case the edge does not exist or it doesn't contain the target type, add it
            Some(_) | None => {
                schema.set_edgetype_unchecked(
                    parsed.source_type,
                    parsed.edge_name,
                    parsed.target_type,
                );
                Ok(NewSchemaAdded)
            }
        }
    } else {
        Ok(NotASchema)
    }
}

/// Used during item creation, completes the item by fields that are not possible to set by the client.
pub fn add_extra_fields(item: &mut CreateItem, cli: &CliOptions, pod_owner: &str) -> Result<()> {
    if item._type == "PluginRun" {
        // New plugin instance is going to be created, get free port for it's webserver
        // and store it in the DB
        let webserver_port = plugin_run::get_free_port(cli);
        item.fields
            .insert(WEBSERVER_PORT.to_string(), json!(webserver_port));

        let container_id = plugin_run::get_container_id(item, cli, pod_owner)?;
        item.fields
            .insert(CONTAINER_ID.to_string(), json!(container_id));

        let webserver_url = plugin_run::resolve_webserver_url(item, cli)?;
        item.fields
            .insert(WEBSERVER_URL.to_string(), json!(webserver_url));
    }

    Ok(())
}

#[derive(Default)]
pub struct AfterAction {
    pub items_to_delete: HashSet<String>,
}

pub fn actions_after_item_create(
    conn: &mut DBConnection,
    id: &str,
    owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<()> {
    let tx_res = in_read_transaction(conn, |tx| {
        let schema = database_api::get_schema(tx)?;

        execute_actions_after_item_create(tx, &schema, id, owner, cli, database_key)
    });

    match tx_res {
        Ok(actions) => {
            // Write TX might be expensive to execute, therefore,
            // firstly check if there is anything to modify
            if !actions.items_to_delete.is_empty() {
                in_write_transaction(conn, |tx| {
                    for del_item in actions.items_to_delete {
                        // Report the error, but do not break the loop
                        if let Err(e) = internal_api::delete_item_by_id(tx, &del_item) {
                            error!(
                                "Removing item {del_item} after item creation failed, reason {e}"
                            );
                        }
                    }

                    Ok(())
                })?;
            }
        }

        Err(e) => error!("Failed upon item triggering, reason {e}"),
    }

    Ok(())
}
/// After item is in the DB do appropriate actions.
/// Invariants:
///  1. This can be called only after successful commit of the write TX. Meaning all new data is visible to all readers.
///  2. Code inside this function cannot do write/update to fields, otherwise it introduces a datarace with readers
///     spawned as a result of this call. Again data will be visible for POD, since its in the middle of the TX,
///     but not outside the TX, as long as it's not committed. When that happens is unknown, hence no assumptions can be made.
/// Returns actions that has to be done in separate transaction.
#[allow(clippy::too_many_arguments)]
fn execute_actions_after_item_create(
    tx: &Tx,
    schema: &Schema,
    item_created_id: &str,
    pod_owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<AfterAction> {
    let item = get_item_tx(tx, schema, item_created_id)?
        .pop()
        .ok_or(Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("Could not find item in the db {}", item_created_id),
        })?;

    let item_base: ItemBase = serde_json::from_value(item.clone())?;

    match item_base._type.as_ref() {
        "PluginRun" => {
            let plugin: PluginRunItem = serde_json::from_value(item)?;
            plugin_run::run_plugin_container(
                tx,
                schema,
                item_created_id,
                &plugin,
                pod_owner,
                database_key,
                cli,
            )
            .map(|_| AfterAction::default())
        }
        _ => plugin_trigger::on_new_item(tx, schema, &item_base, &item, cli),
    }
}

#[cfg(test)]
mod tests {
    use super::add_item_as_schema_opt;
    use crate::api_model::CreateItem;

    use crate::{
        database_api::tests::new_conn,
        error::{Error, Result},
        schema::{Schema, SchemaPropertyType},
        triggers::SchemaAdditionChange,
    };
    use serde_json::json;
    use warp::http::StatusCode;

    #[test]
    fn my_test() -> Result<()> {
        let mut conn = new_conn();
        let tx = conn.transaction().unwrap();

        let mut schema = Schema::empty();
        schema.set_type_unchecked(
            "Person".to_string(),
            "age".to_string(),
            SchemaPropertyType::Integer,
        );

        let json = json!({
            "type": "Something"
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert_eq!(result, Ok(SchemaAdditionChange::NotASchema));

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Integer",
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert_eq!(
            result,
            Err(Error {
                code: StatusCode::INTERNAL_SERVER_ERROR,
                msg: "Failed to find property Person.age".to_string()
            })
        );

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "agility",
            "valueType": "Integer",
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        assert_eq!(schema.all_properties_len(), 1);
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert_eq!(schema.all_properties_len(), 2);
        assert_eq!(
            schema.property("agility"),
            Some(SchemaPropertyType::Integer).as_ref()
        );
        assert_eq!(result, Ok(SchemaAdditionChange::NewSchemaAdded));

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Text",
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert!(result.is_err(), "result should be an error {:?}", result);

        Ok(())
    }
}
