use crate::warp_endpoints;
use libpod::{
    api_model::PodOwner, command_line_interface, command_line_interface::CliOptions,
    db_connection_manager::DBConnectionManager, error::Result, internal_api,
};

use std::{
    net::{IpAddr, SocketAddr},
    ops::Deref,
    path::PathBuf,
    str::FromStr,
    sync::Arc,
};
use tracing::{error, info, warn};
use tracing_subscriber::{filter::EnvFilter, reload::Handle};
use warp::{
    http,
    http::{
        header::{HeaderMap, HeaderValue},
        status::StatusCode,
    },
    hyper::body::Bytes,
    reply::Response,
    Filter, Reply,
};
/// Start web framework with specified APIs.
pub async fn run_server<S: 'static>(cli_options: CliOptions, trace_handle: Handle<EnvFilter, S>) {
    let package_name = env!("CARGO_PKG_NAME").to_uppercase();
    info!("Starting {} HTTP server", package_name);

    let mut headers = HeaderMap::new();
    headers.insert("Access-Control-Allow-Origin", HeaderValue::from_static("*"));
    let headers = warp::reply::with::headers(headers);

    let items_api = warp::path("v4")
        .and(warp::body::content_length_limit(5 * 1024 * 1024))
        .and(warp::post());
    let file_api = warp::path("v4")
        .and(warp::body::content_length_limit(500 * 1024 * 1024))
        .and(warp::post());
    let log_api = warp::path("v4")
        .and(warp::body::content_length_limit(5 * 1024 * 1024))
        .and(warp::post());
    let oauth_api = warp::path("v4")
        .and(warp::body::content_length_limit(5 * 1024 * 1024))
        .and(warp::post());

    let with_cli = {
        let cli_arc = Arc::new(cli_options.clone());
        move || {
            let cli = cli_arc.clone();
            warp::any().map(move || cli.clone())
        }
    };

    let with_db = {
        let db_conn_mgr = DBConnectionManager::new();

        move || {
            let db_conn_mgr_clone = db_conn_mgr.clone();
            warp::any().map(move || db_conn_mgr_clone.clone())
        }
    };

    let with_trace_reload_handler = move || warp::any().map(move || trace_handle.clone());

    let version = warp::path("version")
        .and(warp::path::end())
        .and(warp::get())
        .map(internal_api::get_project_version)
        .boxed();

    let create_item = items_api
        .and(warp::path!(PodOwner / "create_item"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .and(with_cli())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager, cli: Arc<CliOptions>| async move {
                warp_endpoints::create_item(owner, db, body, cli).await
            },
        )
        .boxed();

    let oauth1_request_token =
        oauth_api
            .and(warp::path!(PodOwner / "oauth1_request_token"))
            .and(warp::path::end())
            .and(warp::body::bytes())
            .then(|_: PodOwner, body: Bytes| async move {
                warp_endpoints::oauth1_request_token(body).await
            })
            .boxed();

    let oauth1_access_token =
        oauth_api
            .and(warp::path!(PodOwner / "oauth1_access_token"))
            .and(warp::path::end())
            .and(warp::body::bytes())
            .then(|_: PodOwner, body: Bytes| async move {
                warp_endpoints::oauth1_access_token(body).await
            })
            .boxed();

    let oauth2_auth_url = oauth_api
        .and(warp::path!(PodOwner / "oauth2" / "auth_url"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .then(|owner: PodOwner, body: Bytes| async move {
            warp_endpoints::oauth2_auth_url(owner, body).await
        })
        .boxed();

    let oauth2_authorize = oauth_api
        .and(warp::path!(PodOwner / "oauth2" / "authorize"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .and(with_cli())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager, cli: Arc<CliOptions>| async move {
                warp_endpoints::oauth2_authorize(owner, body, db, cli.deref()).await
            },
        )
        .boxed();

    let oauth2_access_token = oauth_api
        .and(warp::path!(PodOwner / "oauth2" / "access_token"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::oauth2_access_token(owner, body, db).await
            },
        )
        .boxed();

    let get_item = items_api
        .and(warp::path!(PodOwner / "get_item"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::get_item(owner, db, body).await
            },
        )
        .boxed();

    let update_item = items_api
        .and(warp::path!(PodOwner / "update_item"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::update_item(owner, db, body).await
            },
        )
        .boxed();

    let get_edges = items_api
        .and(warp::path!(PodOwner / "get_edges"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::get_edges(owner, db, body).await
            },
        )
        .boxed();

    let create_edge = items_api
        .and(warp::path!(PodOwner / "create_edge"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::create_edge(owner, db, body).await
            },
        )
        .boxed();

    let delete_edge_by_source_target = items_api
        .and(warp::path!(PodOwner / "delete_edge_by_source_target"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::delete_edge_by_source_target(owner, db, body).await
            },
        )
        .boxed();

    let bulk_action = items_api
        .and(warp::path!(PodOwner / "bulk"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .and(with_cli())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager, cli: Arc<CliOptions>| async move {
                warp_endpoints::bulk(owner, body, db, cli).await
            },
        )
        .boxed();

    let delete_item = items_api
        .and(warp::path!(PodOwner / "delete_item"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::delete_item(owner, db, body).await
            },
        )
        .boxed();

    let search = items_api
        .and(warp::path!(PodOwner / "search"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::search(owner, db, body).await
            },
        )
        .boxed();

    let graphql = items_api
        .and(warp::path!(PodOwner / "graphql"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::graphql(owner, db, body).await
            },
        )
        .boxed();

    let upload_file = file_api
        .and(warp::path!(PodOwner / "upload_file" / String / String))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner,
             database_key: String,
             expected_sha256: String,
             body: Bytes,
             db: DBConnectionManager| async move {
                warp_endpoints::upload_file(owner, db, database_key, expected_sha256, body).await
            },
        )
        .boxed();

    let upload_file_b = file_api
        .and(warp::path!(PodOwner / "upload_file_b" / String / String))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner,
             auth_json: String,
             expected_sha256: String,
             body: Bytes,
             db: DBConnectionManager| async move {
                warp_endpoints::upload_file_b(owner, db, auth_json, expected_sha256, body).await
            },
        )
        .boxed();

    let get_file = file_api
        .and(warp::path!(PodOwner / "get_file"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager| async move {
                warp_endpoints::get_file(owner, db, body).await
            },
        )
        .boxed();

    let get_log = log_api
        .and(warp::path!(PodOwner / "get_pluginrun_log"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .and(with_cli())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager, cli: Arc<CliOptions>| async move {
                warp_endpoints::get_logs(owner, db, body, cli).await
            },
        )
        .boxed();

    let send_email = items_api
        .and(warp::path!(PodOwner / "send_email"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .and(with_cli())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager, cli: Arc<CliOptions>| async move {
                warp_endpoints::send_email(owner, db, body, cli).await
            },
        )
        .boxed();

    let trigger_status = items_api
        .and(warp::path!(PodOwner / "trigger" / "status"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_db())
        .and(with_cli())
        .then(
            |owner: PodOwner, body: Bytes, db: DBConnectionManager, cli: Arc<CliOptions>| async move {
                let result =
                    warp_endpoints::trigger_status(owner, db, body, cli).await;
                let result = result.map(|()| warp::reply::json(&serde_json::json!({})));
                respond_with_result(result)
            },
        )
        .boxed();

    let trace_filter = warp::post()
        .and(warp::path("trace_filter"))
        .and(warp::path::end())
        .and(warp::body::bytes())
        .and(with_trace_reload_handler())
        .then(
            |body: Bytes, trace_handler: Handle<EnvFilter, S>| async move {
                warp_endpoints::trace_filter(body, trace_handler).await
            },
        )
        .boxed();

    let origin_request = warp::options()
        .and(warp::header::<String>("origin"))
        .map(move |_origin| {
            let builder = http::response::Response::builder()
                .status(StatusCode::OK)
                .header("access-control-allow-methods", "HEAD, GET, POST, PUT")
                .header(
                    "access-control-allow-headers",
                    "Origin, X-Requested-With, Content-Type, Accept",
                )
                .header("access-control-max-age", "300");
            builder
                .header("vary", "origin")
                .body("".to_string())
                .unwrap()
        })
        .boxed();

    let always_enabled_filters = version
        .with(&headers)
        .or(create_item.with(&headers))
        .boxed();

    let sensitive_filters = warp::any().and_then(|| async move {
        if command_line_interface::PARSED.shared_server {
            Ok(warp::reply::with_status("", StatusCode::NOT_FOUND).into_response())
        } else {
            Err(warp::reject::not_found()) // reject in order to pass to filters below
        }
    });
    let sensitive_filters = sensitive_filters
        .or(get_item.with(&headers))
        .or(bulk_action.with(&headers))
        .or(update_item.with(&headers))
        .or(delete_item.with(&headers))
        .or(search.with(&headers))
        .or(graphql.with(&headers))
        .or(get_edges.with(&headers))
        .or(create_edge.with(&headers))
        .or(delete_edge_by_source_target.with(&headers))
        .or(upload_file.with(&headers))
        .or(upload_file_b.with(&headers))
        .or(get_file.with(&headers))
        .or(get_log.with(&headers))
        .or(send_email.with(&headers))
        .or(trigger_status.with(&headers))
        .or(origin_request.with(&headers))
        .or(oauth1_request_token.with(&headers))
        .or(oauth1_access_token.with(&headers))
        .or(oauth2_auth_url.with(&headers))
        .or(oauth2_authorize.with(&headers))
        .or(oauth2_access_token.with(&headers))
        .or(trace_filter.with(&headers))
        .boxed();

    let not_found = warp::any().map(|| {
        warp::reply::with_status("Endpoint not found", StatusCode::NOT_FOUND).into_response()
    });

    let filters = always_enabled_filters.or(sensitive_filters).or(not_found);

    if cli_options.non_tls || cli_options.insecure_non_tls.is_some() {
        let ip = if let Some(ip) = cli_options.insecure_non_tls {
            if ip.is_loopback() {
                info!(
                    "You seem to use --insecure-non-tls option with a loopback address. \
                    It is recommended to instead use --non-tls CLI option \
                    for clarity and simplicity."
                );
            } else if check_public_ip(&ip) {
                warn!(
                    "The server is asked to run on a public IP {} without https encryption. \
                    This is discouraged as an intermediary (even your router on a local network) \
                    could spoof the traffic sent to the server and do a MiTM attack.\
                    Please consider using Pod with https encryption, \
                    or run it on a non-public network.",
                    ip
                );
            };
            ip
        } else {
            IpAddr::from([127, 0, 0, 1])
        };
        let socket = SocketAddr::new(ip, cli_options.port);
        warp::serve(filters).run(socket).await
    } else {
        let cert_path = &cli_options.tls_pub_crt;
        let key_path = &cli_options.tls_priv_key;
        if PathBuf::from_str(cert_path)
            .map(|p| !p.exists())
            .unwrap_or(true)
        {
            error!("Certificate public key {} not found", cert_path);
            std::process::exit(1)
        };
        if PathBuf::from_str(key_path)
            .map(|p| !p.exists())
            .unwrap_or(true)
        {
            error!("Certificate private key {} not found", cert_path);
            std::process::exit(1)
        };
        let socket = SocketAddr::new(IpAddr::from([0, 0, 0, 0]), cli_options.port);
        warp::serve(filters)
            .tls()
            .cert_path(cert_path)
            .key_path(key_path)
            .run(socket)
            .await;
    }
}

fn respond_with_result<T: Reply>(result: Result<T>) -> Response {
    match result {
        Err(err) => {
            let code = err.code.as_str();
            let code_canon = err.code.canonical_reason().unwrap_or("");
            info!(
                "Returning HTTP failure {} {}: {}",
                code, code_canon, &err.msg
            );
            let msg = format!("Failure: {}", err.msg);
            warp::reply::with_status(msg, err.code).into_response()
        }
        Ok(t) => t.into_response(),
    }
}

fn check_public_ip(addr: &IpAddr) -> bool {
    match addr {
        IpAddr::V4(v4) => v4.is_private() || v4.is_loopback() || v4.is_link_local(),
        IpAddr::V6(v6) if v6.is_loopback() => true,
        // https://en.wikipedia.org/wiki/Unique_local_address
        // Implementation copied from `v6.is_unique_local()`,
        // which is not yet stabilized in Rust
        IpAddr::V6(v6) if (v6.segments()[0] & 0xfe00) == 0xfc00 => true,
        // https://en.wikipedia.org/wiki/Link-local_address
        // Implementation copied from `v6.is_unicast_link_local()`,
        // which is not yet stabilized in Rust
        IpAddr::V6(v6) if (v6.segments()[0] & 0xffc0) == 0xfe80 => true,
        _ => false,
    }
}
