use lazy_static::lazy_static;
use libpod::{
    api_model::{
        AuthKey, Bulk, BulkResponse, CreateEdge, CreateItem, DeleteEdgeBySourceTarget, GetEdges,
        GetFile, Oauth2AccessTokenRequest, Oauth2AccessTokenResponse, Oauth2AuthUrlRequest,
        Oauth2AuthorizeTokenRequest, OauthAccessTokenPayload, OauthRequestTokenPayload,
        PayloadWrapper, PodOwner, Search, SendEmail, TraceRequest, UpdateItem,
    },
    command_line_interface,
    command_line_interface::CliOptions,
    database_api::{self, in_read_transaction, in_write_transaction},
    db_connection_manager::{DBConnection, DBConnectionManager},
    email,
    error::{Error, Result},
    file_api, internal_api, oauth1_api,
    oauth2_api::{self},
    plugin_auth_crypto,
    plugin_auth_crypto::DatabaseKey,
    plugin_run, triggers,
};

use serde_json::Value;
use sha2::{
    digest::generic_array::{typenum::U32, GenericArray},
    Digest, Sha256,
};
use std::{
    collections::{HashMap, HashSet},
    ops::Deref,
    sync::{atomic::AtomicU16, Arc},
};
use tokio::task;
use tracing::{error, info};
use tracing::{instrument, warn};
use tracing_subscriber::{reload::Handle, EnvFilter};

use warp::{http::status::StatusCode, hyper::body::Bytes, reply::Response, Reply};

//
// Items API:
//
#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn get_item(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _get_item(owner, db, body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _get_item(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<Vec<Value>> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<String> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;

    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_read_transaction(&mut conn, |tx| {
        let schema = database_api::get_schema(tx)?;
        internal_api::get_item_tx(tx, &schema, &payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn create_item(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Response {
    let result = _create_item(owner, db, body, cli).await;

    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _create_item(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Result<String> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<CreateItem> = serde_path_to_error::deserialize(body)?;

    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;

    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    let owner = owner;
    // Commit the item to the DB
    let item_id = in_write_transaction(&mut conn, |tx| {
        let mut schema = database_api::get_schema(tx)?;

        internal_api::create_item_tx(tx, &mut schema, payload, &owner, cli.deref())
    });

    if let Ok(id) = &item_id {
        // After transaction commit, the item is visible to all readers, now it's safe to call triggers
        triggers::actions_after_item_create(&mut conn, id, &owner, cli.deref(), &database_key)?;
    }

    item_id
}

#[instrument(skip(body))]
pub async fn oauth1_request_token(body: Bytes) -> Response {
    let result = _oauth1_request_token(body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _oauth1_request_token(body: Bytes) -> Result<Value> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<OauthRequestTokenPayload> = serde_path_to_error::deserialize(body)?;
    let payload = body.payload;
    oauth1_api::oauth1_request_token(payload.callback_url, payload.service).await
}

#[instrument(skip(body))]
pub async fn oauth1_access_token(body: Bytes) -> Response {
    let result = _oauth1_access_token(body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _oauth1_access_token(body: Bytes) -> Result<Value> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<OauthAccessTokenPayload> = serde_path_to_error::deserialize(body)?;
    let payload = body.payload;
    oauth1_api::oauth1_access_token(
        payload.oauth_token,
        payload.oauth_verifier,
        payload.oauth_token_secret,
    )
    .await
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn oauth2_auth_url(owner: PodOwner, body: Bytes) -> Response {
    let result = _oauth2_auth_url(body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _oauth2_auth_url(body: Bytes) -> Result<String> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<Oauth2AuthUrlRequest> = serde_path_to_error::deserialize(body)?;
    let payload = body.payload;

    oauth2_api::auth_url(payload)
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn oauth2_authorize(
    owner: PodOwner,
    body: Bytes,
    db: DBConnectionManager,
    cli: &CliOptions,
) -> Response {
    let result = _oauth2_authorize(owner, body, db, cli).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _oauth2_authorize(
    owner: PodOwner,
    body: Bytes,
    db: DBConnectionManager,
    cli: &CliOptions,
) -> Result<Oauth2AccessTokenResponse> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<Oauth2AuthorizeTokenRequest> = serde_path_to_error::deserialize(body)?;
    let payload = body.payload;

    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;

    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_write_transaction(&mut conn, |tx| {
        let mut schema = database_api::get_schema(tx)?;

        oauth2_api::authorize(tx, &mut schema, &owner, cli, payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn oauth2_access_token(
    owner: PodOwner,
    body: Bytes,
    db: DBConnectionManager,
) -> Response {
    let result = _oauth2_access_token(owner, body, db).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _oauth2_access_token(
    owner: PodOwner,
    body: Bytes,
    db: DBConnectionManager,
) -> Result<Oauth2AccessTokenResponse> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<Oauth2AccessTokenRequest> = serde_path_to_error::deserialize(body)?;

    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;

    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;
    let payload = body.payload;

    in_write_transaction(&mut conn, |tx| {
        let mut schema = database_api::get_schema(tx)?;

        oauth2_api::access_token(tx, &mut schema, payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn update_item(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _update_item(owner, db, body).await;
    let result = result.map(|()| warp::reply::json(&serde_json::json!({})));
    respond_with_result(result)
}

async fn _update_item(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<()> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<UpdateItem> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_write_transaction(&mut conn, |tx| {
        let schema = database_api::get_schema(tx)?;
        internal_api::update_item_tx(tx, &schema, &payload.id, payload.fields)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn bulk(
    owner: PodOwner,
    body: Bytes,
    db: DBConnectionManager,
    cli: Arc<CliOptions>,
) -> Response {
    let result = _bulk(owner, db, body, cli).await;
    let result = result.map(|value| warp::reply::json(&value));

    respond_with_result(result)
}

async fn _bulk(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Result<BulkResponse> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<Bulk> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    let res = in_write_transaction(&mut conn, |tx| {
        let mut schema = database_api::get_schema(tx)?;
        internal_api::bulk_tx(tx, &mut schema, payload, &cli, &owner)
    });

    if let Ok(bulk_res) = &res {
        // After transaction commit, the created items are visible to all readers, now it's safe to call triggers
        for id in bulk_res.create_items.iter() {
            triggers::actions_after_item_create(&mut conn, id, &owner, cli.deref(), &database_key)?;
        }
    }

    res
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn delete_item(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _delete_item(owner, db, body).await;
    let result = result.map(|()| warp::reply::json(&serde_json::json!({})));
    respond_with_result(result)
}

async fn _delete_item(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<()> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<String> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;
    in_write_transaction(&mut conn, |tx| {
        internal_api::delete_item_by_id(tx, &payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn create_edge(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _create_edge(owner, db, body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _create_edge(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<String> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<CreateEdge> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_write_transaction(&mut conn, |tx| {
        let mut schema = database_api::get_schema(tx)?;
        internal_api::create_edge(tx, &mut schema, payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn delete_edge_by_source_target(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
) -> Response {
    let result = _delete_edge_by_source_target(owner, db, body).await;
    let result = result.map(|()| warp::reply::json(&serde_json::json!({})));
    respond_with_result(result)
}

async fn _delete_edge_by_source_target(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
) -> Result<()> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<DeleteEdgeBySourceTarget> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_write_transaction(&mut conn, |tx| {
        internal_api::delete_edge_by_source_target(tx, payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn get_edges(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _get_edges(owner, db, body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}
async fn _get_edges(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<Vec<Value>> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<GetEdges> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_read_transaction(&mut conn, |tx| {
        let schema = database_api::get_schema(tx)?;
        internal_api::get_edges(tx, payload, &schema)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn search(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _search(owner, db, body).await;

    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

async fn _search(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<Vec<Value>> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<Search> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_read_transaction(&mut conn, |tx| {
        let schema = database_api::get_schema(tx)?;
        internal_api::search(tx, &schema, payload)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn graphql(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _graphql(owner, db, body).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

pub async fn _graphql(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
) -> Result<HashMap<String, Vec<Value>>> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<String> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_read_transaction(&mut conn, |tx| {
        let schema = database_api::get_schema(tx)?;
        internal_api::graphql(tx, &schema, payload)
    })
}
//
// Files API:
//
#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn upload_file(
    owner: PodOwner,
    db: DBConnectionManager,
    database_key: String,
    expected_sha256: String,
    body: Bytes,
) -> Response {
    let result = _upload_file(owner, db, database_key, expected_sha256, body).await;
    respond_with_result(result.map(|()| warp::reply::json(&serde_json::json!({}))))
}

async fn _upload_file(
    owner: PodOwner,
    db: DBConnectionManager,
    database_key: String,
    expected_sha256: String,
    body: Bytes,
) -> Result<()> {
    let database_key = DatabaseKey::from(database_key).map(Arc::new)?;

    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_write_transaction(&mut conn, |tx| {
        file_api::upload_file(tx, &owner, &expected_sha256, &body)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn upload_file_b(
    owner: PodOwner,
    db: DBConnectionManager,
    auth_json: String,
    expected_sha256: String,
    body: Bytes,
) -> Response {
    let result = _upload_file_b(owner, db, auth_json, expected_sha256, body).await;
    respond_with_result(result.map(|()| warp::reply::json(&serde_json::json!({}))))
}

async fn _upload_file_b(
    owner: PodOwner,
    db: DBConnectionManager,
    auth_json: String,
    expected_sha256: String,
    body: Bytes,
) -> Result<()> {
    let auth_json = percent_encoding::percent_decode_str(&auth_json)
        .decode_utf8()
        .map_err(|err| Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("Failed to decode UTF-8 authorization JSON, {}", err),
        })?;
    let auth: AuthKey = serde_json::from_str(&auth_json)?;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_write_transaction(&mut conn, |tx| {
        file_api::upload_file(tx, &owner, &expected_sha256, &body)
    })
}

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn get_file(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Response {
    let result = _get_file(owner, db, body).await;
    respond_with_result(result)
}

async fn _get_file(owner: PodOwner, db: DBConnectionManager, body: Bytes) -> Result<Vec<u8>> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<GetFile> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_read_transaction(&mut conn, |tx| {
        file_api::get_file(tx, &owner, &payload.sha256)
    })
}

//
// Logs API
//

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn get_logs(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Response {
    let result = _get_logs(owner, db, body, cli).await;
    let result = result.map(|result| warp::reply::json(&result));
    respond_with_result(result)
}

pub async fn _get_logs(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Result<Value> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<String> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    in_read_transaction(&mut conn, move |tx| {
        plugin_run::get_logs(tx, &payload, &cli)
    })
}

//
// Email API
//

#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub async fn send_email(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Response {
    let result = _send_email(owner, db, body, cli).await;
    let result = result.map(|()| warp::reply::json(&serde_json::json!({})));
    respond_with_result(result)
}

async fn _send_email(
    owner: PodOwner,
    db: DBConnectionManager,
    body: Bytes,
    cli: Arc<CliOptions>,
) -> Result<()> {
    let body = &mut serde_json::Deserializer::from_slice(body.deref());
    let body: PayloadWrapper<SendEmail> = serde_path_to_error::deserialize(body)?;
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let _conn = check_owner_and_initialize_db(&owner, db, &database_key).await?;

    task::spawn_blocking(move || email::send_email(payload, &cli)).await?
}

//
// Trigger status
//
#[instrument(fields(uid=trace_uid(), %owner), skip_all)]
pub(crate) async fn trigger_status(
    owner: PodOwner,
    _db: DBConnectionManager,
    body: Bytes,
    _cli: Arc<CliOptions>,
) -> Result<()> {
    let body = serde_json::from_slice::<Value>(&body);

    info!("Trigger status: {body:#?}");
    // TODO: TBD what to do with the notification

    Ok(())
}

#[instrument(skip_all)]
pub async fn trace_filter<S>(body: Bytes, trace_handler: Handle<EnvFilter, S>) -> Response {
    let result = _trace_filter(body, trace_handler);

    let result = result.map(|()| warp::reply::json(&serde_json::json!("ok")));
    respond_with_result(result)
}

pub fn _trace_filter<S>(body: Bytes, trace_handler: Handle<EnvFilter, S>) -> Result<()> {
    const SECRET_HASH: &str = "E40C1223ED2AB3C68BFC6F738EE817839F8C3E5155C2E8D67968F39B62293E93";

    let body: TraceRequest = serde_json::from_slice(body.as_ref())?;

    let mut hasher = Sha256::new();
    hasher.update(body.secret);
    let res = hasher.finalize();

    if SECRET_HASH != hex::encode_upper(res) {
        return Err(Error {
            code: StatusCode::UNAUTHORIZED,
            msg: "unauthorized".to_string(),
        });
    }

    warn!("Changing trace filter to {}", body.filter);

    let new_filter = body
        .filter
        .parse::<tracing_subscriber::filter::EnvFilter>()
        .map_err(|e| Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("{e}"),
        })?;

    trace_handler.reload(new_filter).map_err(|e| Error {
        code: StatusCode::INTERNAL_SERVER_ERROR,
        msg: format!("{e}"),
    })
}

//
// helper functions:
//

fn auth_to_database_key(auth: AuthKey) -> Result<DatabaseKey> {
    match auth {
        AuthKey::ClientAuth(c) => DatabaseKey::from(c.database_key),
        AuthKey::PluginAuth(p) => plugin_auth_crypto::extract_database_key(&p),
    }
}

/// Two methods combined into one to prevent creating a database connection without owner checks.
/// As additional failsafe to the fact that non-owners don't have the database key.
async fn check_owner_and_initialize_db(
    owner: &str,
    db: DBConnectionManager,
    database_key: &DatabaseKey,
) -> Result<DBConnection> {
    check_owner(owner)?;
    db.initialize_db(owner, database_key).await
}

fn allowed_owner_hashes_fn(cli_options: &CliOptions) -> HashSet<Vec<u8>> {
    let owners: &str = &cli_options.owners;
    let mut result = HashSet::new();
    for owner in owners.split(',') {
        let hexed = hex::decode(owner).unwrap_or_else(|err| {
            error!(
                "Pod owners are invalid, failed to decode hex {}, {}",
                owner, err
            );
            std::process::exit(1);
        });
        result.insert(hexed);
    }
    result
}

lazy_static! {
    static ref ALLOWED_OWNER_HASHES: HashSet<Vec<u8>> =
        allowed_owner_hashes_fn(&command_line_interface::PARSED);
}

fn hash_of_hex(hex_string: &str) -> Result<GenericArray<u8, U32>> {
    let hex_string: Vec<u8> = hex::decode(hex_string)?;
    let mut hash = sha2::Sha256::new();
    hash.update(&hex_string);
    Ok(hash.finalize())
}

fn check_owner(possible_owner: &str) -> Result<()> {
    if command_line_interface::PARSED.owners == "ANY" {
        return Ok(());
    };
    let possible_hash = hash_of_hex(possible_owner)?;
    if ALLOWED_OWNER_HASHES.contains(possible_hash.deref()) {
        Ok(())
    } else {
        info!(
            "Denying unexpected owner {} with hash {}",
            possible_owner,
            hex::encode(possible_hash)
        );
        Err(Error {
            code: StatusCode::FORBIDDEN,
            msg: "Unexpected owner".to_string(),
        })
    }
}

fn respond_with_result<T: Reply>(result: Result<T>) -> Response {
    match result {
        Err(err) => {
            let code = err.code.as_str();
            let code_canon = err.code.canonical_reason().unwrap_or("");
            info!(
                "Returning HTTP failure {} {}: {}",
                code, code_canon, &err.msg
            );
            let msg = format!("Failure: {}", err.msg);
            warp::reply::with_status(msg, err.code).into_response()
        }
        Ok(t) => t.into_response(),
    }
}

/// Get unique identifier for given trace span
fn trace_uid() -> u16 {
    static UID: AtomicU16 = AtomicU16::new(0);

    // It safely wraps around during overflow
    UID.fetch_add(1, std::sync::atomic::Ordering::Relaxed)
}
