// TODO: Clippy has a bug reporting unused fields and methods of PodClient.
#![allow(dead_code)]
use libpod::api_model::{AuthKey, PayloadWrapper};
use reqwest::Response;
use serde::Serialize;
use std::{fmt::Debug, time::Duration};
use tokio::{
    io::{AsyncBufReadExt, BufReader},
    process::{ChildStderr, ChildStdout},
    sync::mpsc::{UnboundedReceiver, UnboundedSender},
    time::timeout,
};
use tracing::{debug, info, instrument};

pub struct PodClient {
    database_key: String,
    owner_key: String,
    pod_url: String,
    pub client: reqwest::Client,
}

/// HTTP client for POD instance
impl PodClient {
    pub fn new(database_key: &str, owner_key: &str, pod_url: &str) -> Self {
        Self {
            database_key: database_key.to_string(),
            owner_key: owner_key.to_string(),
            pod_url: pod_url.to_string(),
            client: reqwest::Client::new(),
        }
    }

    /// Convenient function to do POST request with given payload to the POD
    pub async fn post_to<T>(&self, payload: T, endpoint: &str) -> Response
    where
        T: Serialize + Debug,
    {
        debug!("POST: {endpoint}, body {payload:#?}");

        let body = PayloadWrapper {
            auth: AuthKey::ClientAuth(libpod::api_model::ClientAuth {
                database_key: self.database_key.clone(),
            }),
            payload,
        };

        self.client
            .post(&format!(
                "{}/v4/{}/{endpoint}",
                self.pod_url, self.owner_key
            ))
            .json(&body)
            .send()
            .await
            .unwrap()
    }

    pub async fn get(&self, endpoint: &str) -> Response {
        debug!("GET {endpoint}");
        self.client.get(endpoint).send().await.unwrap()
    }
}

/// Task for log collection from the POD. Streams must be consumed continuously to avoid process buffer
/// saturation and a deadlock. The function displays the POD logs immediately to the console, useful for
/// the developer during debugging, also sends them to the stream that is later used by the tests
/// for analysis.
pub fn start_log_collection(
    std_out: ChildStdout,
    std_err: ChildStderr,
) -> UnboundedReceiver<String> {
    let (tx, rx) = tokio::sync::mpsc::unbounded_channel();
    tokio::spawn(log_collector(std_out, std_err, tx));

    rx
}

#[instrument(name = "", skip_all)]
async fn log_collector(std_out: ChildStdout, std_err: ChildStderr, tx: UnboundedSender<String>) {
    let mut out_reader = BufReader::new(std_out).lines();
    let mut err_reader = BufReader::new(std_err).lines();

    loop {
        tokio::select! {
          Ok(Some(line)) = out_reader.next_line() => {
            info!("STDOUT: {line}");
            if let Err(_e) = tx.send(line) {
                // No more receivers, safe to exit
                break;
            }
          },
          Ok(Some(line)) = err_reader.next_line() => {
            info!("STDERR: {line}");
            if let Err(_e) = tx.send(line) {
                // No more receivers, safe to exit
                break;
            }
          },
          else => {
            break;
          }
        }
    }
}

/// Consumes the current buffer of logs in order to find a line containing given pattern.
/// Returns that line on success, None otherwise.
pub async fn find_in_logs(
    pod_logs: &mut UnboundedReceiver<String>,
    sub_string: &str,
    wait_for: Duration,
) -> Option<String> {
    timeout(wait_for, async {
        while let Some(line) = pod_logs.recv().await {
            if line.contains(sub_string) {
                return Some(line);
            }
        }

        None
    })
    .await
    .unwrap_or(None)
}
