// TODO: Clippy has a bug reporting unused fields and methods that are used across integration tests
#![allow(dead_code)]

use super::pod_client::find_in_logs;
use super::pod_client::PodClient;
use super::test_data::TestData;
use bollard::container::ListContainersOptions;
use libpod::{
    api_model::{Bulk, CreateItem, Search},
    json_utils,
};
use serde_json::Value;
use std::{collections::HashMap, time::Duration};
use tokio::time::timeout;

/// Add complete definition for Plugin and PluginRun types
pub async fn migrate_plugin_definition(client: &PodClient) {
    let plugin_def = serde_json::from_str::<Bulk>(include_str!("plugin_migration.json")).unwrap();

    let resp = client.post_to(plugin_def, "bulk").await;
    assert_eq!(resp.status(), reqwest::StatusCode::OK);
}

pub async fn deploy_plugin(client: &PodClient) {
    let payload = serde_json::from_str::<Bulk>(
        r#"
    {
        "createItems":[
            {
                "type":"Plugin",
                "id":"b8d3cd38-1a0f-492d-8cbb-39dde5b5e090",

                "gitProjectId":234,
                "config":"{\"item_type\":\"Message\",\"model_name\":\"my_plugin_model\",\"model_version\":\"0.1\",\"isMock\":true}",
                "configJson":"[\n  {\n    \"name\": \"item_type\",\n    \"display\": \"Item Type\",\n    \"data_type\": \"Text\",\n    \"type\": \"textbox\",\n    \"default\": \"Message\",\n    \"optional\": true\n  },\n  {\n    \"name\": \"item_service\",\n    \"display\": \"Item Service\",\n    \"data_type\": \"Text\",\n    \"type\": \"textbox\",\n    \"default\": null,\n    \"optional\": true\n  },\n  {\n    \"name\": \"model_name\",\n    \"display\": \"Model Name\",\n    \"data_type\": \"Text\",\n    \"type\": \"textbox\",\n    \"default\": \"my_plugin_model\",\n    \"optional\": true\n  },\n  {\n    \"name\": \"model_version\",\n    \"display\": \"Model Version\",\n    \"data_type\": \"Text\",\n    \"type\": \"textbox\",\n    \"default\": \"0.1\",\n    \"optional\": true\n  },\n  {\n    \"name\": \"isMock\",\n    \"display\": \"Ismock\",\n    \"data_type\": \"Bool\",\n    \"type\": \"textbox\",\n    \"default\": true,\n    \"optional\": true\n  }\n]",
                "containerImage":"gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest",
                "name":"plugin to trigger",
                "pluginDescription":"A transformer based sentiment analyis plugin",
                "pluginModule":"plugin_to_trigger.plugin",
                "pluginName":"ClassifierPlugin",
                "pluginType":"app"
            },
            {
                "type":"Edge",
                "id":"5c6ec341-182e-4e8f-9c31-5745a829b4b6"
            },

            {
                "type": "Project",
                "id": "e9ad0695-e874-4f01-86e0-08339bf1ee8a"
            }
        ],
        "createEdges":[
            {
                "_source":"e9ad0695-e874-4f01-86e0-08339bf1ee8a",
                "_target":"b8d3cd38-1a0f-492d-8cbb-39dde5b5e090",
                "_self":"5c6ec341-182e-4e8f-9c31-5745a829b4b6",
                "_name":"labellingPlugin"
            }
        ]
    }
    "#,
    ).unwrap();

    let resp = client.post_to(payload, "bulk").await;
    assert_eq!(resp.status(), reqwest::StatusCode::OK);
}

pub async fn start_plugin(ctx: &mut TestData) -> String {
    let plugin_run_id = "4fa2094c-51d6-4874-a135-9017fcdedf16".to_string();

    let payload = serde_json::from_str::<Bulk>(
        r#"
    {
        "createItems": [
          {
            "type": "PluginRun",
            "id": "4fa2094c-51d6-4874-a135-9017fcdedf16",

            "containerImage": "gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest",
            "pluginModule": "plugin_to_trigger.plugin",
            "pluginName": "ClassifierPlugin",
            "status": "idle",
            "targetItemId": "4fa2094c-51d6-4874-a135-9017fcdedf16"
          },
          {
            "type": "Edge",
            "id": "f6632a28-312e-486e-a790-dd3b415098e3"
          }
        ],
        "createEdges": [
          {
            "_source": "4fa2094c-51d6-4874-a135-9017fcdedf16",
            "_target": "b8d3cd38-1a0f-492d-8cbb-39dde5b5e090",
            "_self": "f6632a28-312e-486e-a790-dd3b415098e3",
            "_name": "plugin"
          }
        ]
      }"#,
    )
    .unwrap();

    let resp = ctx.pod_client.post_to(payload, "bulk").await;
    assert_eq!(resp.status(), reqwest::StatusCode::OK);

    // Starting container takes time...
    // It can be even worse, if happens for first time, and image must be pulled from the repo
    // Check with docker API if container exists
    let containers = timeout(Duration::from_secs(90), async {
        let mut containers = vec![];

        while containers.is_empty() {
            containers = ctx
                .docker_api
                .list_containers(Some(ListContainersOptions {
                    filters: HashMap::from([
                        ("status", vec!["running"]),
                        ("name", vec![&ctx.owner_key]),
                    ]),
                    ..Default::default()
                }))
                .await
                .expect("Failed to list running containers");
        }

        containers
    })
    .await
    .expect("Timeout while waiting for running container");

    assert_eq!(
        containers.len(),
        1,
        "Expected one container running for the owner: {}, containers: {containers:#?}",
        ctx.owner_key
    );

    // Plugin execution should be visible in the logs
    assert!(find_in_logs(
        &mut ctx.pod_logs,
        "Loading model...",
        Duration::from_secs(30)
    )
    .await
    .is_some());

    assert!(find_in_logs(
        &mut ctx.pod_logs,
        "Start predicting...",
        Duration::from_secs(30)
    )
    .await
    .is_some());

    assert!(
        find_in_logs(&mut ctx.pod_logs, "Done", Duration::from_secs(30))
            .await
            .is_some()
    );

    plugin_run_id
}

pub async fn delete_plugin(pod_client: &PodClient) {
    let plugin_id = "4fa2094c-51d6-4874-a135-9017fcdedf16";

    let resp = pod_client.post_to(plugin_id, "delete_item").await;
    assert_eq!(resp.status(), reqwest::StatusCode::OK);

    let plugin_item = get_db_item(plugin_id, pod_client).await;

    let plugin_deleted = json_utils::get_property_value(&plugin_item, "deleted")
        .expect("delete plugin get property")
        .as_bool()
        .expect("delete plugin parse to bool");

    assert!(plugin_deleted);
}

pub async fn register_trigger(client: &PodClient) -> String {
    let trigger_id = "0a4a06a4-fb6c-4ff0-b12a-f64fb1f84e41".to_string();

    let payload = serde_json::from_str::<Bulk>(
        r#"{
    "createItems": [
      {
        "type": "Trigger",
        "id": "0a4a06a4-fb6c-4ff0-b12a-f64fb1f84e41",

        "action": "v1/item/trigger",
        "pluginRunId": "4fa2094c-51d6-4874-a135-9017fcdedf16",
        "triggerOn": "Message",
        "filterCreatedAfter": 1655170839992,
        "filterCreatedAfterPropertyName": "dateSent"
      },
      {
        "type": "Edge",
        "id": "68c13d4a-b8cc-49ff-ae56-cf4d0d9cf89b"
      }
    ],
    "createEdges": [
      {
        "_source": "0a4a06a4-fb6c-4ff0-b12a-f64fb1f84e41",
        "_target": "4fa2094c-51d6-4874-a135-9017fcdedf16",
        "_self": "68c13d4a-b8cc-49ff-ae56-cf4d0d9cf89b",
        "_name": "trigger"
      }
    ]
  }"#,
    )
    .unwrap();

    let resp = client.post_to(payload, "bulk").await;
    assert_eq!(resp.status(), reqwest::StatusCode::OK);

    trigger_id
}

pub async fn send_fake_msg(client: &PodClient) {
    let payload = serde_json::from_str::<CreateItem>(
        r#"
        {
            "type": "Message",
            "content": "this is a fake content",
            "dateSent": 1655170839993
        }
        "#,
    )
    .unwrap();

    let resp = client.post_to(payload, "create_item").await;
    assert_eq!(resp.status(), reqwest::StatusCode::OK);
}

pub async fn get_db_item(item_id: &str, client: &PodClient) -> Value {
    let payload = Search {
        id: Some(item_id.to_string()),
        limit: 10,
        ..Default::default()
    };
    let resp = client.post_to(payload, "search").await;

    resp.json::<Vec<Value>>()
        .await
        .expect("search response")
        .pop()
        .expect("search result is empty")
}
